from opentrons.protocol_api.protocol_context import ProtocolContext


def run(protocol_context: ProtocolContext) -> None:
    tip_rack = protocol_context.load_labware("opentrons_96_tiprack_10ul", "3")
    plate = protocol_context.load_labware("biorad_96_wellplate_200ul_pcr", "1")

    pipette = protocol_context.load_instrument(
        "p10_single", "left", tip_racks=[tip_rack]
    )

    pipette.transfer(5, plate.wells("A1"), plate.wells("B1"))



import asyncio
import logging
import os
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional

from opentrons_shared_data.errors.codes import ErrorCodes
from opentrons_shared_data.errors.exceptions import CommunicationError
from opentrons_shared_data.robot.types import RobotType, RobotTypeEnum
from opentrons_shared_data.util import StrEnum

from opentrons.config import (
    ARCHITECTURE,
    IS_ROBOT,
    SystemArchitecture,
    get_opentrons_path,
)
from opentrons.protocol_engine.resources.camera_provider import (
    CameraError,
    CameraProvider,
    CameraSettings,
    ImageParameters,
)
from opentrons.system import ffmpeg

log = logging.getLogger(__name__)

# Default System Cameras
DEFAULT_SYSTEM_CAMERA = "/dev/ot_system_camera"

# Default Preview Image Filename
PREVIEW_IMAGE = "preview_image.jpeg"

# Stream Globals
DEFAULT_CONF_FILE = (
    "/lib/systemd/system/opentrons-live-stream/opentrons-live-stream.env"
)
STREAM_CONF_FILE_KEYS = [
    "BOOT_ID",
    "STATUS",
    "SOURCE",
    "RESOLUTION",
    "FRAMERATE",
    "BITRATE",
]

# Camera Parameter Globals
RESOLUTION_MIN = (320, 240)
RESOLUTION_MAX = (7680, 4320)
RESOLUTION_DEFAULT = (1920, 1080)
ZOOM_MIN = 1.0
ZOOM_MAX = 2.0
ZOOM_DEFAULT = 1.0
CONTRAST_MIN = 0.0
CONTRAST_MAX = 2.0
CONTRAST_DEFAULT = 1.0
BRIGHTNESS_MIN = -128
BRIGHTNESS_MAX = 128
BRIGHTNESS_DEFAULT = 0
SATURATION_MIN = 0.0
SATURATION_MAX = 2.0
SATURATION_DEFAULT = 1.0


class StreamConfigurationKeys(StrEnum):
    """The Configuration Key Types."""

    BOOT_ID = "BOOT_ID"
    STATUS = "STATUS"
    SOURCE = "SOURCE"
    RESOLUTION = "RESOLUTION"
    FRAMERATE = "FRAMERATE"
    BITRATE = "BITRATE"


class CameraException(CommunicationError):
    def __init__(self, message: str, system_error: str) -> None:
        super().__init__(
            ErrorCodes.COMMUNICATION_ERROR,
            message,
            {"internal-error-message": system_error},
        )


async def take_picture(filename: Path) -> None:
    """Legacy method to take a picture and save it to filename

    :param filename: Name of file to save picture to
    :param loop: optional loop to use
    :return: None
    :raises: CameraException
    """
    try:
        os.remove(filename)
    except OSError:
        pass

    if ARCHITECTURE == SystemArchitecture.YOCTO:
        cmd = f"v4l2-ctl --device /dev/ot_system_camera --set-fmt-video=width=1280,height=720,pixelformat=MJPG --stream-mmap --stream-to={str(filename)} --stream-count=1"
    elif ARCHITECTURE == SystemArchitecture.BUILDROOT:
        cmd = f"ffmpeg -f video4linux2 -s 640x480 -i /dev/video0 -ss 0:0:1 -frames 1 {str(filename)}"
    else:  # HOST
        cmd = f'ffmpeg -f avfoundation -framerate 1  -s 640x480  -i "0" -ss 0:0:1 -frames 1 {str(filename)}'

    proc = await asyncio.create_subprocess_shell(
        cmd,
        stderr=asyncio.subprocess.PIPE,
    )

    res = await proc.stderr.read()  # type: ignore
    res = res.decode().strip()
    await proc.wait()

    if proc.returncode != 0:
        raise CameraException("Failed to communicate with camera", res)
    if not filename.exists():
        raise CameraException("Failed to save image", "")


def get_stream_configuration_filepath() -> Path:
    """Return the file path to the Opentrons Live Stream Configuration file."""
    filepath = get_opentrons_path("live_stream_environment_file")
    if IS_ROBOT and not os.path.exists(filepath):
        # If the dynamic configuration file doesn't exist make it using our defaults file
        with open(DEFAULT_CONF_FILE, "r") as default_config:
            content = default_config.read()
        with open(filepath, "w") as new_config_file:
            new_config_file.write(content)
    return filepath


def robot_supports_livestream(robot_type: RobotType) -> bool:
    """Validate whether or not robot supports live streaming service."""
    robot = RobotTypeEnum.robot_literal_to_enum(robot_type)
    if robot == RobotTypeEnum.OT2:
        # If we are on an OT-2 we do not support live streams
        return False
    return True


async def update_live_stream_status(
    robot_type: RobotType,
    stream_status: bool,
    camera_provider: CameraProvider,
    override_settings: Optional[CameraSettings] = None,
) -> None:
    """Update and handle a change in the Opentrons Live Stream status."""
    if not IS_ROBOT or robot_supports_livestream(robot_type) is False:
        # If we are not on a robot we simply no-op updating the stream
        return None

    contents = load_stream_configuration_file_data()
    if contents is None:
        log.error("Opentrons Live Stream Configuration file cannot be updated.")
        return None

    # Validate the stream status
    if override_settings is not None:
        camera_enable_settings = override_settings
    else:
        camera_enable_settings = await camera_provider.get_camera_settings()
    status = "OFF"
    if (
        stream_status
        and camera_enable_settings.cameraEnabled
        and camera_enable_settings.liveStreamEnabled
    ):
        # Check to see if the camera device is available
        raw_device = str(contents["SOURCE"])
        if not os.path.exists(raw_device):
            log.error(
                f"Opentrons Live Stream cannot sample the camera. No video device found with device path: {raw_device}"
            )
        # Enable the stream
        status = "ON"
    # Overwrite the contents
    contents["BOOT_ID"] = get_boot_id()
    contents["STATUS"] = status
    write_stream_configuration_file_data(contents)
    await restart_live_stream(robot_type)


async def stop_live_stream(robot_type: RobotType) -> None:
    """Attempt to stop the Opentrons Live Stream service."""
    if robot_supports_livestream(robot_type) is False:
        # No-op on OT-2 since we don't have a live stream service there
        return None

    command = ["systemctl", "stop", "opentrons-live-stream"]
    subprocess = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await subprocess.communicate()
    if subprocess.returncode == 0:
        log.info("Stopped the opentrons-live-stream service.")
    else:
        log.error(
            f"Failed to stop opentrons-live-stream, returncode: {subprocess.returncode}, stdout: {stdout.decode()}, stderr: {stderr.decode()}"
        )


async def restart_live_stream(robot_type: RobotType) -> None:
    """Attempt to restart the Opentrons Live Stream service."""
    if robot_supports_livestream(robot_type) is False:
        # No-op on OT-2 since we don't have a live stream service there
        return None

    command = ["systemctl", "restart", "opentrons-live-stream"]
    subprocess = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await subprocess.communicate()
    if subprocess.returncode == 0:
        log.info("Restarted opentrons-live-stream service.")
    else:
        log.error(
            f"Failed to restart opentrons-live-stream, returncode: {subprocess.returncode}, stdout: {stdout.decode()}, stderr: {stderr.decode()}"
        )


def load_stream_configuration_file_data() -> dict[str, str] | None:
    """Load the Opentrons Live Stream Conf file and return parsed data or None if an error occurs."""
    src = get_stream_configuration_filepath()
    if not src.exists():
        log.error(f"Opentrons Live Stream configuration file not found: {src}")
        return None
    with src.open("rb") as fd:
        try:
            return parse_stream_configuration_file_data(fd.read())
        except Exception as e:
            log.error(
                f"Opentrons Live Stream status update file parsing failed with: {e}"
            )
    return None


def parse_stream_configuration_file_data(data: bytes) -> Dict[str, str] | None:
    """
    Parse a collect of bytes for Opentrons Live Stream Configuration data and return a dictionary of
    results keyed by configuration constants. Returns None if an error occurred during parsing.
    """
    contents: Dict[str, str] = {
        key.decode("utf-8"): val.decode("utf-8")
        for key, val in [line.split(b"=") for line in data.split(b"\n") if b"=" in line]
    }

    enum_stream_keys = {stream_key.value for stream_key in StreamConfigurationKeys}
    if sorted(list(contents.keys())) != sorted(enum_stream_keys):
        log.error(
            "Opentrons Live Stream Configuration file data is incorrect or missing."
        )
        # We don't want to write bad or incomplete data to the file
        return None

    # Migrate old camera default file data to new uniform default
    if contents[StreamConfigurationKeys.SOURCE] == "NONE":
        contents[StreamConfigurationKeys.SOURCE] = DEFAULT_SYSTEM_CAMERA
    return contents


def write_stream_configuration_file_data(data: Dict[str, str]) -> None:
    src = get_stream_configuration_filepath()
    if not src.exists():
        log.error(f"Opentrons Live Stream configuration file not found: {src}")
        return None

    enum_stream_keys = {stream_key.value for stream_key in StreamConfigurationKeys}
    if sorted(list(data.keys())) != sorted(enum_stream_keys):
        log.error(
            "Data provided to write is not compatible with Opentrons Live Stream Configuration file."
        )
        return None

    with src.open("w") as fd:
        file_lines = [
            f"{StreamConfigurationKeys.BOOT_ID}={data[StreamConfigurationKeys.BOOT_ID]}\n",
            f"{StreamConfigurationKeys.STATUS}={data[StreamConfigurationKeys.STATUS]}\n",
            f"{StreamConfigurationKeys.SOURCE}={data[StreamConfigurationKeys.SOURCE]}\n",
            f"{StreamConfigurationKeys.RESOLUTION}={data[StreamConfigurationKeys.RESOLUTION]}\n",
            f"{StreamConfigurationKeys.FRAMERATE}={data[StreamConfigurationKeys.FRAMERATE]}\n",
            f"{StreamConfigurationKeys.BITRATE}={data[StreamConfigurationKeys.BITRATE]}\n",
        ]
        fd.writelines(file_lines)


async def image_capture(  # noqa: C901
    robot_type: RobotType, parameters: ImageParameters
) -> bytes | CameraError:
    """Process an Image Capture request with a Camera utilizing a given set of parameters."""
    camera = DEFAULT_SYSTEM_CAMERA

    # We must always validate the camera exists
    if not os.path.exists(camera):
        return CameraError(
            message=f"No video device found with device path {camera}", code=None
        )

    if parameters.zoom is not None and (
        parameters.zoom < ZOOM_MIN or parameters.zoom > ZOOM_MAX
    ):
        potential_invalid_param = "Zoom"
    elif parameters.contrast is not None and (
        parameters.contrast < CONTRAST_MIN or parameters.contrast > CONTRAST_MAX
    ):
        potential_invalid_param = "Contrast"
    elif parameters.brightness is not None and (
        parameters.brightness < BRIGHTNESS_MIN or parameters.brightness > BRIGHTNESS_MAX
    ):
        potential_invalid_param = "Brightness"
    elif parameters.saturation is not None and (
        parameters.saturation < SATURATION_MIN or parameters.saturation > SATURATION_MAX
    ):
        potential_invalid_param = "Saturation"
    elif parameters.resolution is not None and (
        parameters.resolution[0] < RESOLUTION_MIN[0]
        or parameters.resolution[1] < RESOLUTION_MIN[1]
        or parameters.resolution[0] > RESOLUTION_MAX[0]
        or parameters.resolution[1] > RESOLUTION_MAX[1]
    ):
        potential_invalid_param = "Resolution"
    else:
        potential_invalid_param = None

    if potential_invalid_param is not None:
        return CameraError(
            message=f"{potential_invalid_param} parameter is outside the boundaries allowed for image capture.",
            code="IMAGE_SETTINGS",
        )
    try:
        # Always stop the live stream service to ensure the Camera is always free when attempting an image capture
        await stop_live_stream(robot_type)

        zoom = parameters.zoom if parameters.zoom is not None else ZOOM_DEFAULT
        contrast = (
            parameters.contrast if parameters.contrast is not None else CONTRAST_DEFAULT
        )
        brightness = (
            parameters.brightness
            if parameters.brightness is not None
            else BRIGHTNESS_DEFAULT
        )
        saturation = (
            parameters.saturation
            if parameters.saturation is not None
            else SATURATION_DEFAULT
        )
        resolution = (
            parameters.resolution
            if parameters.resolution is not None
            else RESOLUTION_DEFAULT
        )

        result = await ffmpeg.ffmpeg_capture_image_bytes(
            robot_type=robot_type,
            resolution=resolution,
            camera=camera,
            zoom=zoom,
            pan=parameters.pan if parameters.pan is not None else (0, 0),
            contrast=contrast,
            brightness=brightness,
            saturation=saturation,
        )
    except Exception:
        result = CameraError(
            message="Exception occured during execution of system image capture.",
            code=None,
        )
    finally:
        # Restart the live stream service
        await restart_live_stream(robot_type)
    return result


@lru_cache(maxsize=1)
def get_boot_id() -> str:
    if IS_ROBOT:
        return Path("/proc/sys/kernel/random/boot_id").read_text().strip()
    else:
        return "SIMULATED_BOOT_ID"


def camera_exists() -> bool:
    """Validate whether or not the camera device exists."""
    return os.path.exists(DEFAULT_SYSTEM_CAMERA)
    # todo(chb, 2025-11-10): Eventually when we support multiple cameras this should accept a camera parameter to check for



import asyncio
import sys
from typing import AsyncGenerator, List

import anyio
import pytest

from .build_module import build_module
from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.hardware_control.modules import Thermocycler
from opentrons.hardware_control.modules.types import TemperatureStatus, ThermocyclerStep


@pytest.fixture
async def thermocycler(
    emulation_app: None,
    emulator_settings: Settings,
    execution_manager: ExecutionManager,
    poll_interval_seconds: float,
) -> AsyncGenerator[Thermocycler, None]:
    """Return a Thermocycler test subject."""
    module = await build_module(
        Thermocycler,
        port=emulator_settings.thermocycler_proxy.driver_port,
        execution_manager=execution_manager,
        poll_interval_seconds=poll_interval_seconds,
    )
    try:
        yield module
    finally:
        await module.cleanup()


def test_device_info(thermocycler: Thermocycler) -> None:
    """It should have device info."""
    assert {
        "model": "thermocyclerModuleV2",
        "serial": "thermocycler_emulator",
        "version": "v1.1.0",
    } == thermocycler.device_info


async def test_lid_status(thermocycler: Thermocycler) -> None:
    """It should run open and close lid."""
    assert thermocycler.lid_status == "open"

    await thermocycler.close()
    assert thermocycler.lid_status == "closed"

    await thermocycler.open()
    assert thermocycler.lid_status == "open"


async def test_lid_temperature(thermocycler: Thermocycler) -> None:
    """It should change lid temperature."""
    await thermocycler.set_lid_temperature(temperature=50)
    assert thermocycler.lid_target == 50

    await thermocycler.set_lid_temperature(temperature=40)
    assert thermocycler.lid_target == 40

    await thermocycler.deactivate_lid()
    assert thermocycler.lid_target is None


async def test_forcible_deactivate(
    thermocycler: Thermocycler, execution_manager: ExecutionManager
) -> None:
    """Can override wait_for_is_running."""
    await execution_manager.pause()
    await thermocycler.deactivate(must_be_running=False)
    await thermocycler.deactivate_block(must_be_running=False)
    await thermocycler.deactivate_lid(must_be_running=False)


async def test_plate_temperature(thermocycler: Thermocycler) -> None:
    """It should change  plate temperature."""
    await thermocycler.set_temperature(temperature=52, hold_time_seconds=10)
    assert thermocycler.temperature == 52

    await thermocycler.set_temperature(temperature=55, hold_time_seconds=None)
    assert thermocycler.temperature == 55

    await thermocycler.set_temperature(temperature=80, hold_time_seconds=1)
    assert thermocycler.temperature == 80

    await thermocycler.deactivate_block()
    assert thermocycler.target is None


async def test_cycle_temperatures(thermocycler: Thermocycler) -> None:
    """It should cycle the temperature."""
    assert thermocycler.current_cycle_index is None
    assert thermocycler.current_step_index is None
    assert thermocycler.total_cycle_count is None
    assert thermocycler.total_step_count is None

    steps: list[ThermocyclerStep] = [
        {
            "temperature": 70.0,
        },
        {
            "temperature": 60.0,
            "hold_time_minutes": 1.0,
        },
        {
            "temperature": 50.0,
            "hold_time_seconds": 22.0,
        },
    ]
    await thermocycler.cycle_temperatures(steps, repetitions=2)

    # Check that final temperature was reached.
    assert thermocycler.temperature == 50
    # Check that cycle state is correct.
    assert thermocycler.current_cycle_index == 2
    assert thermocycler.current_step_index == 3
    assert thermocycler.total_cycle_count == 2
    assert thermocycler.total_step_count == 3

    # Now deactivate block to check that cycle state is cleared
    await thermocycler.deactivate_block()

    assert thermocycler.current_cycle_index is None
    assert thermocycler.current_step_index is None
    assert thermocycler.total_cycle_count is None
    assert thermocycler.total_step_count is None


async def test_wait_for_temperatures(thermocycler: Thermocycler) -> None:
    """It should wait for temperature and be at holding status"""
    await thermocycler.set_target_block_temperature(40.0)
    await thermocycler.set_target_lid_temperature(50.0)
    await thermocycler.wait_for_block_target()
    await thermocycler.wait_for_lid_target()
    assert thermocycler.status == TemperatureStatus.HOLDING
    assert thermocycler.lid_temp_status == TemperatureStatus.HOLDING


# TODO(mm, 2022-07-01): This test is a flakiness hazard because it's sensitive to
# timing and async task scheduling. Move to an isolated unit test, instead
@pytest.mark.xfail(
    condition=(sys.platform == "darwin"), reason="Timing flakiness on macOS"
)
async def test_cycle_cannot_be_interrupted_by_pause(
    poll_interval_seconds: int,
    thermocycler: Thermocycler,
    execution_manager: ExecutionManager,
) -> None:
    """If the execution manager is paused, it should not suspend the current cycle.

    https://github.com/Opentrons/opentrons/issues/5496
    """
    temp_1 = 40.0
    temp_2 = 50.0
    final_temp = 60.0

    # Start at a known temperature.
    await thermocycler.set_temperature(temperature=temp_1)

    # Oscillate between two temperatures for a while
    # and then end on a third temperature.
    #
    # This list must be long enough that we can reliably target a pause somewhere
    # in the middle of it, but not so long that makes the test take a disruptively
    # long time.
    steps: list[ThermocyclerStep] = [
        {"temperature": temp_1, "hold_time_seconds": poll_interval_seconds * 2},
        {"temperature": temp_2, "hold_time_seconds": poll_interval_seconds * 2},
        {"temperature": temp_1, "hold_time_seconds": poll_interval_seconds * 2},
        {"temperature": temp_2, "hold_time_seconds": poll_interval_seconds * 2},
        {"temperature": final_temp},
    ]

    # 1-based index, arbitrarily chosen somewhere in the middle.
    step_to_pause_after = 2

    # Start cycling through the steps.
    cycle_temperatures_task = asyncio.create_task(
        thermocycler.cycle_temperatures(steps=steps, repetitions=1)
    )

    # Wait until we reach step_to_pause_after or any of the steps that follow it.
    while (
        thermocycler.current_step_index is None
        or thermocycler.current_step_index < step_to_pause_after
    ):
        await asyncio.sleep(0)

    # For this test to be meaningful, this task needs to have woken up before the steps
    # have completed. Double-check that this is actually the case.
    # Note that current_step_index is 1-based.
    assert thermocycler.current_step_index <= thermocycler.total_step_count  # type: ignore[operator]
    assert thermocycler.temperature != final_temp

    # Issue a pause now that we're in the middle of the steps.
    await execution_manager.pause()

    # All of the steps should complete, despite the pause. The pause should be ignored.
    #
    # If the subject has a bug where the pause actually takes effect,
    # this would stall forever. The anyio.fail_after() turns that into a TimeoutError
    # for nicer reporting.
    with anyio.fail_after(10):
        await cycle_temperatures_task
        assert thermocycler.temperature == final_temp


async def test_cycle_can_be_blocked_by_preexisting_pause(
    thermocycler: Thermocycler,
    execution_manager: ExecutionManager,
) -> None:
    """A pre-existing pause should block a new cycle from starting.

    Not to be confused with a pause that happens in the middle of a cycle.
    """
    temp_1 = 20.0
    temp_2 = 30.0

    # Start at a known temperature.
    await thermocycler.set_temperature(temperature=temp_1)

    # Pause before we begin a cycle.
    await execution_manager.pause()

    # Assert that starting a cycle blocks for at least 0.5 seconds,
    # as an approximation of asserting that it blocks forever.
    with pytest.raises(TimeoutError):
        with anyio.fail_after(0.5):
            steps: List[ThermocyclerStep] = [{"temperature": temp_2}]
            await thermocycler.cycle_temperatures(steps=steps, repetitions=1)

    # Assert that the cycle didn't have any effect.
    assert thermocycler.current_step_index is None
    assert thermocycler.temperature == temp_1


async def test_cycle_can_be_cancelled(
    thermocycler: Thermocycler,
    execution_manager: ExecutionManager,
) -> None:
    """A cycle should be cancellable (even though it isn't pausable)."""
    steps: List[ThermocyclerStep] = [
        {"temperature": 20.0, "hold_time_minutes": 99999},
    ]

    cycle_temperatures_task = asyncio.create_task(
        thermocycler.cycle_temperatures(steps=steps, repetitions=1)
    )

    # Wait until we're somewhere in the middle of the steps.
    while thermocycler.current_step_index is None:
        await asyncio.sleep(0)

    # Issue a cancellation now that we're in the middle of the steps.
    await execution_manager.cancel()

    # Assert that the cancellation takes effect promptly, within 0.5 seconds.
    with anyio.fail_after(0.5):
        with pytest.raises(asyncio.CancelledError):
            await cycle_temperatures_task



"""Tests for Protocol API point calculation."""

from typing import Tuple

import pytest

from opentrons.protocol_api.core.engine import point_calculations as subject
from opentrons.types import Point


def test_get_relative_offset() -> None:
    """It should get a relative offset given a point, size in x, y, z, and x, y, z ratios."""
    result = subject.get_relative_offset(
        point=Point(1.0, 2.0, 3.0),
        size=(4, 8, 16),
        x_ratio=0.5,
        y_ratio=0.25,
        z_ratio=0.125,
    )

    assert result == Point(2.0, 3.0, 4.0)


@pytest.mark.parametrize(
    argnames=["rect1", "rect2", "expected_result"],
    argvalues=[
        (  # One rectangle inside the other
            (Point(236, 170, 0), Point(391, 375, 0)),
            (Point(237, 250, 0), Point(388, 294, 0)),
            True,
        ),
        (  # One rectangle inside the other
            (Point(237, 250, 0), Point(388, 294, 0)),
            (Point(236, 170, 0), Point(391, 375, 0)),
            True,
        ),
        (  # Two non-overlapping rectangles
            (Point(236, 170, 0), Point(391, 375, 0)),
            (Point(438, 216, 100), Point(937, 306, 200)),
            False,
        ),
        (  # Two non-overlapping rectangles in 2nd quadrant
            (Point(-438, 216, 100), Point(-937, 306, 200)),
            (Point(-236, 170, 0), Point(-391, 375, 0)),
            False,
        ),
        (  # Overlapping rectangles with one corner of each rectangle overlapping
            (Point(719, 304, 20), Point(970, 370, 20)),
            (Point(438, 216, 100), Point(937, 306, 200)),
            True,
        ),
        (  # Overlapping rectangles with no overlapping corners
            # (think two rectangles making a '+' sign)
            (Point(630, 94, 20), Point(800, 500, 20)),
            (Point(438, 216, 100), Point(937, 306, 200)),
            True,
        ),
    ],
)
def test_are_overlapping_rectangles(
    rect1: Tuple[Point, Point],
    rect2: Tuple[Point, Point],
    expected_result: bool,
) -> None:
    """It should calculate correctly whether the rectangles are overlapping."""
    assert (
        subject.are_overlapping_rectangles(rectangle1=rect1, rectangle2=rect2)
        == expected_result
    )



"""Tests for the ProtocolContext public interface."""

import inspect
from typing import Dict, cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data import liquid_classes
from opentrons_shared_data.labware.types import LabwareDefinition as LabwareDefDict
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    PositionReference,
)
from opentrons_shared_data.liquid_classes.types import TransferPropertiesDict
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.robot.types import RobotType
from tests.opentrons.protocol_api import (
    versions_at_or_above,
    versions_between,
)

from opentrons.hardware_control.modules.types import (
    FlexStackerModuleModel,
    MagneticBlockModel,
    ModuleType,
    TemperatureModuleModel,
)
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import (
    MAX_SUPPORTED_VERSION,
    OFF_DECK,
    Deck,
    InstrumentContext,
    Labware,
    Liquid,
    MagneticBlockContext,
    MagneticModuleContext,
    ModuleContext,
    ProtocolContext,
    TemperatureModuleContext,
)
from opentrons.protocol_api import (
    validation as mock_validation,
)
from opentrons.protocol_api._liquid import LiquidClass
from opentrons.protocol_api.core.common import (
    FlexStackerCore,
    InstrumentCore,
    LabwareCore,
    MagneticBlockCore,
    MagneticModuleCore,
    ProtocolCore,
    TemperatureModuleCore,
)
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocol_api.disposal_locations import TrashBin, WasteChute
from opentrons.protocol_api.tasks import Task
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors import LabwareMovementNotAllowedError
from opentrons.protocols.api_support import instrument as mock_instrument_support
from opentrons.protocols.api_support.deck_type import (
    NoTrashDefinedError,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import (
    APIVersionError,
    RobotTypeError,
    UnsupportedAPIError,
)
from opentrons.types import DeckSlotName, Mount, StagingSlotName


@pytest.fixture(autouse=True)
def _mock_validation_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_instrument_support_module(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    for name, func in inspect.getmembers(mock_instrument_support, inspect.isfunction):
        monkeypatch.setattr(mock_instrument_support, name, decoy.mock(func=func))


@pytest.fixture
def mock_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock implementation core."""
    mock_core = decoy.mock(cls=ProtocolCore)
    mock_fixed_trash = decoy.mock(cls=LabwareCore)
    decoy.when(mock_core.fixed_trash).then_return(mock_fixed_trash)
    decoy.when(mock_fixed_trash.get_name()).then_return("cool trash")
    decoy.when(mock_fixed_trash.get_display_name()).then_return("Cool Trash")
    decoy.when(mock_fixed_trash.get_well_columns()).then_return([])
    return mock_core


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_deck(decoy: Decoy) -> Deck:
    """Get a mock Deck."""
    return decoy.mock(cls=Deck)


@pytest.fixture
def mock_fixed_trash(decoy: Decoy) -> Labware:
    """Get a mock Fixed Trash."""
    return decoy.mock(cls=Labware)


@pytest.fixture
def api_version() -> APIVersion:
    """The API version under test."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def subject(
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_deck: Deck,
    api_version: APIVersion,
    mock_fixed_trash: Labware,
    decoy: Decoy,
) -> ProtocolContext:
    """Get a ProtocolContext test subject with its dependencies mocked out."""
    decoy.when(mock_core_map.get(mock_core.fixed_trash)).then_return(mock_fixed_trash)
    return ProtocolContext(
        api_version=api_version,
        core=mock_core,
        core_map=mock_core_map,
        deck=mock_deck,
    )


def test_legacy_trash_loading(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_fixed_trash: Labware,
    mock_deck: Deck,
) -> None:
    """It should load a trash labware on init on API level 2.15 and below."""
    decoy.when(mock_core_map.get(mock_core.fixed_trash)).then_return(mock_fixed_trash)
    context = ProtocolContext(
        api_version=APIVersion(2, 15),
        core=mock_core,
        core_map=mock_core_map,
        deck=mock_deck,
    )
    assert mock_fixed_trash == context.fixed_trash
    decoy.verify(mock_core.append_disposal_location(mock_fixed_trash))


def test_automatic_ot2_trash_loading(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_deck: Deck,
) -> None:
    """It should load a trash labware on init on API level 2.15 and below."""
    decoy.when(mock_core.robot_type).then_return("OT-2 Standard")
    ProtocolContext(
        api_version=APIVersion(2, 16),
        core=mock_core,
        core_map=mock_core_map,
        deck=mock_deck,
    )
    decoy.verify(mock_core.load_ot2_fixed_trash_bin())


def test_fixed_trash(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ProtocolContext,
) -> None:
    """It should get the fixed trash labware from the core."""
    trash_captor = matchers.Captor()

    assert mock_core.fixed_trash is not None
    decoy.verify(mock_core_map.add(mock_core.fixed_trash, trash_captor), times=1)

    trash = trash_captor.value

    decoy.when(mock_core_map.get(mock_core.fixed_trash)).then_return(trash)
    decoy.when(mock_core.get_disposal_locations()).then_return([trash])
    result = subject.fixed_trash

    assert result is trash
    assert isinstance(result, Labware)
    assert result.name == "cool trash"


def test_deck(subject: ProtocolContext) -> None:
    """It should have a Deck interface."""
    result = subject.deck
    assert isinstance(result, Deck)


@pytest.mark.parametrize("api_version", [APIVersion(2, 20)])
def test_load_instrument_robot_type(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """Non-Flex robot type should raise a ValueError."""
    mock_tip_racks = [decoy.mock(cls=Labware), decoy.mock(cls=Labware)]

    decoy.when(mock_validation.ensure_lowercase_name("Gandalf")).then_return("gandalf")
    decoy.when(mock_validation.ensure_pipette_name("gandalf")).then_return(
        PipetteNameType.P300_SINGLE
    )
    decoy.when(
        mock_validation.ensure_mount_for_pipette(
            "shadowfax", PipetteNameType.P300_SINGLE
        )
    ).then_return(Mount.LEFT)
    decoy.when(mock_core.robot_type).then_return("OT-2 Standard")

    with pytest.raises(RobotTypeError):
        subject.load_instrument(
            instrument_name="Gandalf",
            mount="shadowfax",
            tip_racks=mock_tip_racks,
            liquid_presence_detection=False,
        )


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_load_instrument_api_version(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """Using an API Version prior to 2.20 should raise a APIVersionError."""
    mock_tip_racks = [decoy.mock(cls=Labware), decoy.mock(cls=Labware)]

    decoy.when(mock_validation.ensure_lowercase_name("Gandalf")).then_return("gandalf")
    decoy.when(mock_validation.ensure_pipette_name("gandalf")).then_return(
        PipetteNameType.P300_SINGLE
    )
    decoy.when(
        mock_validation.ensure_mount_for_pipette(
            "shadowfax", PipetteNameType.P300_SINGLE
        )
    ).then_return(Mount.LEFT)
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")

    with pytest.raises(APIVersionError):
        subject.load_instrument(
            instrument_name="Gandalf",
            mount="shadowfax",
            tip_racks=mock_tip_racks,
            liquid_presence_detection=False,
        )


def test_load_instrument(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should create an instrument using its execution core."""
    mock_instrument_core = decoy.mock(cls=InstrumentCore)
    mock_tip_racks = [decoy.mock(cls=Labware), decoy.mock(cls=Labware)]

    decoy.when(mock_validation.ensure_lowercase_name("Gandalf")).then_return("gandalf")
    decoy.when(mock_validation.ensure_pipette_name("gandalf")).then_return(
        PipetteNameType.P300_SINGLE
    )
    decoy.when(
        mock_validation.ensure_mount_for_pipette(
            "shadowfax", PipetteNameType.P300_SINGLE
        )
    ).then_return(Mount.LEFT)

    decoy.when(
        mock_core.load_instrument(
            instrument_name=PipetteNameType.P300_SINGLE,
            mount=Mount.LEFT,
            liquid_presence_detection=False,
        )
    ).then_return(mock_instrument_core)

    decoy.when(mock_instrument_core.get_pipette_name()).then_return("Gandalf the Grey")
    decoy.when(mock_instrument_core.get_model()).then_return("wizard")
    decoy.when(mock_core.get_disposal_locations()).then_raise(
        NoTrashDefinedError("No trash!")
    )

    result = subject.load_instrument(
        instrument_name="Gandalf", mount="shadowfax", tip_racks=mock_tip_racks
    )

    assert isinstance(result, InstrumentContext)
    assert result.name == "Gandalf the Grey"
    assert result.requested_as == "gandalf"
    assert subject.loaded_instruments["left"] is result

    decoy.verify(
        mock_instrument_support.validate_tiprack(
            instrument_name="Gandalf the Grey",
            tip_rack=mock_tip_racks[0],
            log=matchers.Anything(),
        ),
        mock_instrument_support.validate_tiprack(
            instrument_name="Gandalf the Grey",
            tip_rack=mock_tip_racks[1],
            log=matchers.Anything(),
        ),
    )


def test_load_instrument_replace(
    decoy: Decoy, mock_core: ProtocolCore, subject: ProtocolContext
) -> None:
    """It should allow/disallow pipette replacement."""
    mock_instrument_core = decoy.mock(cls=InstrumentCore)

    decoy.when(mock_validation.ensure_lowercase_name(matchers.IsA(str))).then_return(
        "ada"
    )
    decoy.when(mock_validation.ensure_pipette_name(matchers.IsA(str))).then_return(
        PipetteNameType.P300_SINGLE
    )
    decoy.when(
        mock_validation.ensure_mount_for_pipette(
            matchers.IsA(Mount), matchers.IsA(PipetteNameType)
        )
    ).then_return(Mount.RIGHT)
    decoy.when(
        mock_core.load_instrument(
            instrument_name=matchers.IsA(PipetteNameType),
            mount=matchers.IsA(Mount),
            liquid_presence_detection=False,
        )
    ).then_return(mock_instrument_core)
    decoy.when(mock_instrument_core.get_pipette_name()).then_return("Ada Lovelace")
    decoy.when(mock_core.get_disposal_locations()).then_raise(
        NoTrashDefinedError("No trash!")
    )

    pipette_1 = subject.load_instrument(instrument_name="ada", mount=Mount.RIGHT)
    assert subject.loaded_instruments["right"] is pipette_1

    pipette_2 = subject.load_instrument(
        instrument_name="ada", mount=Mount.RIGHT, replace=True
    )
    assert subject.loaded_instruments["right"] is pipette_2

    with pytest.raises(RuntimeError, match="Instrument already present"):
        subject.load_instrument(instrument_name="ada", mount=Mount.RIGHT)


def test_96_channel_pipette_raises_if_another_pipette_attached(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should always raise when loading a 96-channel pipette when another pipette is attached."""
    mock_instrument_core = decoy.mock(cls=InstrumentCore)

    decoy.when(
        mock_validation.ensure_lowercase_name("A Single Channel Name")
    ).then_return("a single channel name")
    decoy.when(
        mock_validation.ensure_pipette_name("a single channel name")
    ).then_return(PipetteNameType.P300_SINGLE)
    decoy.when(
        mock_validation.ensure_mount_for_pipette(
            Mount.RIGHT, PipetteNameType.P300_SINGLE
        )
    ).then_return(Mount.RIGHT)

    decoy.when(
        mock_core.load_instrument(
            instrument_name=PipetteNameType.P300_SINGLE,
            mount=Mount.RIGHT,
            liquid_presence_detection=False,
        )
    ).then_return(mock_instrument_core)

    decoy.when(mock_instrument_core.get_pipette_name()).then_return("ada")

    decoy.when(mock_core.get_disposal_locations()).then_raise(
        NoTrashDefinedError("No trash!")
    )

    pipette_1 = subject.load_instrument(
        instrument_name="A Single Channel Name", mount=Mount.RIGHT
    )
    assert subject.loaded_instruments["right"] is pipette_1

    decoy.when(mock_validation.ensure_lowercase_name("A 96 Channel Name")).then_return(
        "a 96 channel name"
    )
    decoy.when(mock_validation.ensure_pipette_name("a 96 channel name")).then_return(
        PipetteNameType.P1000_96
    )
    decoy.when(
        mock_validation.ensure_mount_for_pipette("shadowfax", PipetteNameType.P1000_96)
    ).then_return(Mount.LEFT)
    decoy.when(
        mock_core.load_instrument(
            instrument_name=PipetteNameType.P1000_96,
            mount=Mount.LEFT,
        )
    ).then_return(mock_instrument_core)

    with pytest.raises(RuntimeError):
        subject.load_instrument(instrument_name="A 96 Channel Name", mount="shadowfax")


def test_load_labware(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create a labware using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LABWARE")).then_return(
        "lowercase_labware"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_5)

    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_labware",
            location=DeckSlotName.SLOT_5,
            label="some_display_name",
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name="UPPERCASE_LABWARE",
        location=42,
        label="some_display_name",
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize(
    "label,sanitized_label", [(7, "7"), (None, None), ("hi", "hi")]
)
def test_load_labware_sanitizes_label(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    label: str | None,  # think of this like a typecast
    sanitized_label: str | None,
    subject: ProtocolContext,
) -> None:
    """It should stringify labels unless they are None."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LABWARE")).then_return(
        "lowercase_labware"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_5)

    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_labware",
            location=DeckSlotName.SLOT_5,
            label=sanitized_label,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name="UPPERCASE_LABWARE",
        location=42,
        label=label,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_load_labware_off_deck(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ProtocolContext,
) -> None:
    """It should load labware off-deck."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LABWARE")).then_return(
        "lowercase_labware"
    )

    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_labware",
            location=OFF_DECK,
            label="some_display_name",
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name="UPPERCASE_LABWARE",
        label="some_display_name",
        namespace="some_namespace",
        version=1337,
        location=OFF_DECK,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_load_labware_off_deck_raises(
    subject: ProtocolContext,
) -> None:
    """It should raise and api error."""
    with pytest.raises(APIVersionError):
        subject.load_labware(
            load_name="UPPERCASE_LABWARE",
            label="some_display_name",
            namespace="some_namespace",
            version=1337,
            location=OFF_DECK,
        )


def test_load_labware_on_staging_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create a labware on a staging slot using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LABWARE")).then_return(
        "lowercase_labware"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(StagingSlotName.SLOT_B4)

    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_labware",
            location=StagingSlotName.SLOT_B4,
            label="some_display_name",
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name="UPPERCASE_LABWARE",
        location=42,
        label="some_display_name",
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_load_labware_from_definition(
    decoy: Decoy,
    mock_core: ProtocolCore,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should be able to load a labware from a definition dictionary."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    labware_definition_dict = cast(LabwareDefDict, {"labwareDef": True})
    labware_load_params = LabwareLoadParams("you", "are", 1337)

    decoy.when(mock_validation.ensure_lowercase_name("are")).then_return("are")
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_1)
    decoy.when(mock_core.add_labware_definition(labware_definition_dict)).then_return(
        labware_load_params
    )

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    decoy.when(
        mock_core.load_labware(
            namespace="you",
            load_name="are",
            version=1337,
            location=DeckSlotName.SLOT_1,
            label="Some Display Name",
        )
    ).then_return(mock_labware_core)

    result = subject.load_labware_from_definition(
        labware_def=labware_definition_dict,
        location=42,
        label="Some Display Name",
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"


def test_load_adapter(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create an adapter using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_ADAPTER")).then_return(
        "lowercase_adapter"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_5)

    decoy.when(
        mock_core.load_adapter(
            load_name="lowercase_adapter",
            location=DeckSlotName.SLOT_5,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_adapter(
        load_name="UPPERCASE_ADAPTER",
        location=42,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


def test_load_adapter_on_staging_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create an adapter on a staging slot using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_ADAPTER")).then_return(
        "lowercase_adapter"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(StagingSlotName.SLOT_B4)

    decoy.when(
        mock_core.load_adapter(
            load_name="lowercase_adapter",
            location=StagingSlotName.SLOT_B4,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_adapter(
        load_name="UPPERCASE_ADAPTER",
        location=42,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 25)])
def test_load_labware_lid_adapter_namespace_version_requires_new_api_version(
    subject: ProtocolContext,
) -> None:
    """Make sure parameters new to apiLevel 2.26 raise, if given in older apiLevels."""
    with pytest.raises(APIVersionError, match="adapter_namespace"):
        subject.load_labware("load_name", "A1", adapter_namespace="foo")

    with pytest.raises(APIVersionError, match="adapter_version"):
        subject.load_labware("load_name", "A1", adapter_version=123)

    with pytest.raises(APIVersionError, match="lid_namespace"):
        subject.load_labware("load_name", "A1", lid_namespace="foo")

    with pytest.raises(APIVersionError, match="lid_version"):
        subject.load_labware("load_name", "A1", lid_version=123)


@pytest.mark.parametrize(
    (
        "api_version",
        "input_adapter_namespace",
        "input_adapter_version",
        "expected_adapter_namespace",
        "expected_adapter_version",
    ),
    [
        *[
            # Old APIVersion: Adapter namespace and version cannot be specified explicitly.
            # Adapter namespace always follows main labware, and adapter version is always None.
            (
                v,
                None,
                None,
                sentinel.input_namespace,
                None,
            )
            for v in versions_between(
                low_inclusive_bound=APIVersion(2, 15),
                high_exclusive_bound=APIVersion(2, 26),
            )
        ],
        *[
            # New APIVersion: Adapter namespace and version are used as-is if specified explicitly.
            (
                v,
                sentinel.input_adapter_namespace,
                sentinel.input_adapter_version,
                sentinel.input_adapter_namespace,
                sentinel.input_adapter_version,
            )
            for v in versions_at_or_above(APIVersion(2, 26))
        ],
        *[
            # New APIVersion: Adapter namespace and version default to None if not provided.
            (
                v,
                None,
                None,
                None,
                None,
            )
            for v in versions_at_or_above(APIVersion(2, 26))
        ],
    ],
)
def test_load_labware_on_adapter(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    input_adapter_namespace: str | None,
    input_adapter_version: int | None,
    expected_adapter_namespace: str | None,
    expected_adapter_version: int | None,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create a labware using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_adapter_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_validation.ensure_lowercase_name(sentinel.input_load_name)
    ).then_return(sentinel.lowercase_input_load_name)

    decoy.when(
        mock_validation.ensure_lowercase_name(sentinel.input_adapter)
    ).then_return(sentinel.lowercase_input_adapter)
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(
            sentinel.input_location, api_version, "OT-3 Standard"
        )
    ).then_return(sentinel.validated_location)
    decoy.when(
        mock_core.load_adapter(
            load_name=sentinel.lowercase_input_adapter,
            location=sentinel.validated_location,
            namespace=expected_adapter_namespace,
            version=expected_adapter_version,
        )
    ).then_return(mock_adapter_core)

    decoy.when(mock_adapter_core.get_well_columns()).then_return([])

    decoy.when(
        mock_core.load_labware(
            load_name=sentinel.lowercase_input_load_name,
            location=mock_adapter_core,
            label="input_label",
            namespace=sentinel.input_namespace,
            version=sentinel.input_version,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name=sentinel.input_load_name,
        location=sentinel.input_location,
        label="input_label",
        namespace=sentinel.input_namespace,
        version=sentinel.input_version,
        adapter=sentinel.input_adapter,
        adapter_namespace=input_adapter_namespace,
        adapter_version=input_adapter_version,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize(
    (
        "api_version",
        "input_lid_namespace",
        "input_lid_version",
        "expected_lid_namespace",
        "expected_lid_version",
    ),
    [
        *[
            # Old APIVersion: Lid namespace and version cannot be specified explicitly.
            # Lid namespace and version always follow main labware.
            (
                v,
                None,
                None,
                sentinel.input_namespace,
                sentinel.input_version,
            )
            for v in versions_between(
                low_inclusive_bound=APIVersion(2, 23),
                high_exclusive_bound=APIVersion(2, 26),
            )
        ],
        *[
            # New APIVersion: Lid namespace and version are used as-is if specified explicitly.
            (
                v,
                sentinel.input_lid_namespace,
                sentinel.input_lid_version,
                sentinel.input_lid_namespace,
                sentinel.input_lid_version,
            )
            for v in versions_at_or_above(APIVersion(2, 26))
        ],
        *[
            # New APIVersion: Lid namespace and version default to None if not provided.
            (
                v,
                None,
                None,
                None,
                None,
            )
            for v in versions_at_or_above(APIVersion(2, 26))
        ],
    ],
)
def test_load_labware_with_lid(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    input_lid_namespace: str | None,
    input_lid_version: int | None,
    expected_lid_namespace: str | None,
    expected_lid_version: int | None,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create a labware with a lid on it using its execution core."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(
        mock_validation.ensure_lowercase_name(sentinel.input_load_name)
    ).then_return(sentinel.lowercase_input_load_name)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(
            sentinel.input_location, api_version, "OT-3 Standard"
        )
    ).then_return(sentinel.validated_location)

    decoy.when(
        mock_core.load_labware(
            load_name=sentinel.lowercase_input_load_name,
            location=sentinel.validated_location,
            label="input_label",
            namespace=sentinel.input_namespace,
            version=sentinel.input_version,
        )
    ).then_return(mock_labware_core)

    decoy.when(mock_labware_core.get_name()).then_return("Full Name")
    decoy.when(mock_labware_core.get_display_name()).then_return("Display Name")
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name=sentinel.input_load_name,
        location=sentinel.input_location,
        label="input_label",
        namespace=sentinel.input_namespace,
        version=sentinel.input_version,
        lid=sentinel.input_lid,
        lid_namespace=input_lid_namespace,
        lid_version=input_lid_version,
    )

    assert isinstance(result, Labware)
    assert result.name == "Full Name"

    decoy.verify(
        mock_core.load_lid(
            # todo(mm, 2025-08-26): We're passing load_name=input_lid directly without lowercasing it,
            # unlike how we lowercase adapter names. Is this a bug?
            load_name=sentinel.input_lid,
            location=mock_labware_core,
            namespace=expected_lid_namespace,
            version=expected_lid_version,
        )
    )

    decoy.verify(mock_core_map.add(mock_labware_core, result), times=1)


@pytest.mark.parametrize("api_version", [APIVersion(2, 23)])
def test_load_lid_stack(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should create a labware with a lid on it using its execution core."""
    mock_lid_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LID")).then_return(
        "lowercase_lid"
    )

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_C1)

    decoy.when(
        mock_core.load_lid_stack(
            load_name="lowercase_lid",
            location=DeckSlotName.SLOT_C1,
            quantity=5,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid_core)

    decoy.when(mock_lid_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid_core.get_display_name()).then_return("")
    decoy.when(mock_lid_core.get_well_columns()).then_return([])

    result = subject.load_lid_stack(
        load_name="UPPERCASE_LID",
        location=42,
        quantity=5,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)
    assert result.name == "STACK_OBJECT"


@pytest.mark.parametrize("api_version", [APIVersion(2, 23)])
def test_move_lids_from_stack(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should move all the lids from a lid stack down to the base slot."""
    mock_lid_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LID")).then_return(
        "lowercase_lid"
    )

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_C1)

    decoy.when(
        mock_core.load_lid_stack(
            load_name="lowercase_lid",
            location=DeckSlotName.SLOT_C1,
            quantity=5,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid_core)

    decoy.when(mock_lid_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid_core.get_display_name()).then_return("")
    decoy.when(mock_lid_core.get_well_columns()).then_return([])

    stack = subject.load_lid_stack(
        load_name="UPPERCASE_LID",
        location=42,
        quantity=5,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(stack, Labware)
    assert stack.name == "STACK_OBJECT"

    for i in range(5):
        subject.move_lid(stack, "D3")

    # Load another labware where the lidstack once was, verifying its engine object is gone
    mock_lw_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LABWARE")).then_return(
        "lowercase_labware"
    )
    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_labware",
            location=DeckSlotName.SLOT_C1,
            label=None,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lw_core)

    decoy.when(mock_lw_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lw_core.get_display_name()).then_return("")
    decoy.when(mock_lw_core.get_well_columns()).then_return([])

    result = subject.load_labware(
        load_name="UPPERCASE_LABWARE",
        location=42,
        label=None,
        namespace="some_namespace",
        version=1337,
    )
    assert isinstance(result, Labware)


@pytest.mark.parametrize("api_version", [APIVersion(2, 23)])
def test_move_lids_from_stack_via_stack_parent(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should move a lid onto an empty riser, create a stack, and then move the lid back off by referencing the riser."""
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_C1)
    mock_riser_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_validation.ensure_lowercase_name("RISER_LABWARE")).then_return(
        "riser_labware"
    )
    decoy.when(
        mock_core.load_labware(
            load_name="riser_labware",
            location=DeckSlotName.SLOT_C1,
            label=None,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_riser_core)

    decoy.when(mock_riser_core.get_name()).then_return("RISER_LABWARE")
    decoy.when(mock_riser_core.get_display_name()).then_return("")
    decoy.when(mock_riser_core.get_well_columns()).then_return([])

    riser_lw = subject.load_labware(
        load_name="RISER_LABWARE",
        location=42,
        label=None,
        namespace="some_namespace",
        version=1337,
    )
    assert isinstance(riser_lw, Labware)

    # Load the lid stack on top of the riser
    mock_lid_core = decoy.mock(cls=LabwareCore)
    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LID")).then_return(
        "lowercase_lid"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_core.load_lid_stack(
            load_name="lowercase_lid",
            location=riser_lw._core,
            quantity=1,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid_core)

    decoy.when(mock_lid_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid_core.get_display_name()).then_return("")
    decoy.when(mock_lid_core.get_well_columns()).then_return([])

    result = subject.load_lid_stack(
        load_name="UPPERCASE_LID",
        location=riser_lw,
        quantity=1,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result, Labware)
    assert result.name == "STACK_OBJECT"

    # Move the lid, by referencing only the riser itself
    subject.move_lid(riser_lw, "D3")

    # Load another lid stack where the lidstack once was, verifying its engine object is gone
    mock_lid_core_2 = decoy.mock(cls=LabwareCore)
    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LID_2")).then_return(
        "lowercase_lid_2"
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_core.load_lid_stack(
            load_name="lowercase_lid_2",
            location=riser_lw._core,
            quantity=1,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid_core_2)

    decoy.when(mock_lid_core_2.get_name()).then_return("STACK_OBJECT_2")
    decoy.when(mock_lid_core_2.get_display_name()).then_return("")
    decoy.when(mock_lid_core_2.get_well_columns()).then_return([])

    result_2 = subject.load_lid_stack(
        load_name="UPPERCASE_LID_2",
        location=riser_lw,
        quantity=1,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(result_2, Labware)
    assert result_2.name == "STACK_OBJECT_2"


@pytest.mark.parametrize("api_version", [APIVersion(2, 22)])
def test_move_labware_lids_old(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """This should use the original loading lids as labware for backwards compatibility."""
    mock_lid_core = decoy.mock(cls=LabwareCore)
    mock_lid2_core = decoy.mock(cls=LabwareCore)
    mock_lid3_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_validation.ensure_lowercase_name("UPPERCASE_LID")).then_return(
        "lowercase_lid"
    )

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_C1)
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(43, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_C2)

    decoy.when(
        mock_core.load_labware(
            load_name="lowercase_lid",
            location=DeckSlotName.SLOT_C1,
            label=None,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid_core)

    decoy.when(mock_lid_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid_core.get_display_name()).then_return("")
    decoy.when(mock_lid_core.get_well_columns()).then_return([])

    lid_1 = subject.load_labware(
        load_name="UPPERCASE_LID",
        location=42,
        label=None,
        namespace="some_namespace",
        version=1337,
    )
    assert isinstance(lid_1, Labware)

    decoy.when(
        lid_1._protocol_core.load_labware(
            load_name="lowercase_lid",
            location=lid_1._core,
            label=None,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid2_core)
    decoy.when(mock_lid2_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid2_core.get_display_name()).then_return("")
    decoy.when(mock_lid2_core.get_well_columns()).then_return([])

    lid_2 = lid_1.load_labware(
        name="lowercase_lid",
        label=None,
        namespace="some_namespace",
        version=1337,
    )
    assert isinstance(lid_2, Labware)

    decoy.when(
        lid_2._protocol_core.load_labware(
            load_name="lowercase_lid",
            location=lid_2._core,
            label=None,
            namespace="some_namespace",
            version=1337,
        )
    ).then_return(mock_lid3_core)
    decoy.when(mock_lid3_core.get_name()).then_return("STACK_OBJECT")
    decoy.when(mock_lid3_core.get_display_name()).then_return("")
    decoy.when(mock_lid3_core.get_well_columns()).then_return([])

    lid_3 = lid_1.load_labware(
        name="lowercase_lid",
        label=None,
        namespace="some_namespace",
        version=1337,
    )

    assert isinstance(lid_3, Labware)

    subject.move_labware(lid_3, 43)
    subject.move_labware(lid_2, lid_3)
    subject.move_labware(lid_1, lid_2)


def test_loaded_labware(
    decoy: Decoy,
    mock_core_map: LoadedCoreMap,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should return a list of all loaded modules."""
    labware_core_4 = decoy.mock(cls=LabwareCore)
    labware_core_6 = decoy.mock(cls=LabwareCore)
    labware_4 = decoy.mock(cls=Labware)
    labware_6 = decoy.mock(cls=Labware)

    decoy.when(mock_core.get_labware_cores()).then_return(
        [labware_core_4, labware_core_6]
    )
    decoy.when(labware_core_4.get_deck_slot()).then_return(DeckSlotName.SLOT_4)
    decoy.when(labware_core_6.get_deck_slot()).then_return(DeckSlotName.SLOT_6)
    decoy.when(mock_core_map.get(labware_core_4)).then_return(labware_4)
    decoy.when(mock_core_map.get(labware_core_6)).then_return(labware_6)

    result = subject.loaded_labwares

    assert result == {4: labware_4, 6: labware_6}


def test_move_labware_to_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should move labware to new slot location."""
    drop_offset = {"x": 4, "y": 5, "z": 6}
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_1)
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )
    decoy.when(
        mock_validation.ensure_valid_labware_offset_vector(drop_offset)
    ).then_return((1, 2, 3))
    subject.move_labware(
        labware=movable_labware,
        new_location=42,
        drop_offset=drop_offset,
    )

    decoy.verify(
        mock_core.move_labware(
            labware_core=mock_labware_core,
            new_location=DeckSlotName.SLOT_1,
            use_gripper=False,
            pause_for_manual_move=True,
            pick_up_offset=None,
            drop_offset=(1, 2, 3),
        )
    )


def test_move_labware_to_staging_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should move labware to new slot location."""
    drop_offset = {"x": 4, "y": 5, "z": 6}
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(StagingSlotName.SLOT_B4)
    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )
    decoy.when(
        mock_validation.ensure_valid_labware_offset_vector(drop_offset)
    ).then_return((1, 2, 3))
    subject.move_labware(
        labware=movable_labware,
        new_location=42,
        drop_offset=drop_offset,
    )

    decoy.verify(
        mock_core.move_labware(
            labware_core=mock_labware_core,
            new_location=StagingSlotName.SLOT_B4,
            use_gripper=False,
            pause_for_manual_move=True,
            pick_up_offset=None,
            drop_offset=(1, 2, 3),
        )
    )


def test_move_labware_to_module(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ProtocolContext,
) -> None:
    """It should move labware to new module location."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    mock_module_core = decoy.mock(cls=TemperatureModuleCore)
    mock_broker = decoy.mock(cls=LegacyBroker)

    decoy.when(mock_labware_core.get_well_columns()).then_return([])
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_A1)
    decoy.when(mock_core.get_labware_on_module(mock_module_core)).then_return(None)
    decoy.when(mock_core_map.get(None)).then_return(None)

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )
    module_location = TemperatureModuleContext(
        core=mock_module_core,
        api_version=MAX_SUPPORTED_VERSION,
        broker=mock_broker,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )

    subject.move_labware(labware=movable_labware, new_location=module_location)
    decoy.verify(
        mock_core.move_labware(
            labware_core=mock_labware_core,
            new_location=mock_module_core,
            use_gripper=False,
            pause_for_manual_move=True,
            pick_up_offset=None,
            drop_offset=None,
        )
    )


def test_move_labware_off_deck(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ProtocolContext,
) -> None:
    """It should move labware off-deck."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )

    subject.move_labware(labware=movable_labware, new_location=OFF_DECK)
    decoy.verify(
        mock_core.move_labware(
            labware_core=mock_labware_core,
            new_location=OFF_DECK,
            use_gripper=False,
            pause_for_manual_move=True,
            pick_up_offset=None,
            drop_offset=None,
        )
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_move_labware_off_deck_raises(
    subject: ProtocolContext,
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
) -> None:
    """It should raise an APIVersionError if using move_labware in an unsupported version."""
    mock_labware_core = decoy.mock(cls=LabwareCore)

    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )

    with pytest.raises(APIVersionError):
        subject.move_labware(labware=movable_labware, new_location=OFF_DECK)


def test_move_labware_to_trash_raises(
    subject: ProtocolContext,
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_engine_client: EngineClient,
) -> None:
    """It should raise an LabwareMovementNotAllowedError if using move_labware to move something that is not a lid to a TrashBin."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    trash_location = TrashBin(
        location=DeckSlotName.SLOT_D3,
        addressable_area_name="moveableTrashD3",
        api_version=MAX_SUPPORTED_VERSION,
        engine_client=mock_engine_client,
    )

    decoy.when(mock_labware_core.get_well_columns()).then_return([])

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )

    with pytest.raises(LabwareMovementNotAllowedError):
        subject.move_labware(labware=movable_labware, new_location=trash_location)


def test_move_lid_to_trash_passes(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    subject: ProtocolContext,
    mock_engine_client: EngineClient,
) -> None:
    """It should move a lid labware into a trashbin successfully."""
    mock_labware_core = decoy.mock(cls=LabwareCore)
    trash_location = TrashBin(
        location=DeckSlotName.SLOT_D3,
        addressable_area_name="moveableTrashD3",
        api_version=MAX_SUPPORTED_VERSION,
        engine_client=mock_engine_client,
    )

    decoy.when(mock_labware_core.get_well_columns()).then_return([])
    decoy.when(mock_labware_core.is_lid()).then_return(True)

    movable_labware = Labware(
        core=mock_labware_core,
        api_version=MAX_SUPPORTED_VERSION,
        protocol_core=mock_core,
        core_map=mock_core_map,
    )

    subject.move_labware(labware=movable_labware, new_location=trash_location)
    decoy.verify(
        mock_core.move_labware(
            labware_core=mock_labware_core,
            new_location=trash_location,
            use_gripper=False,
            pause_for_manual_move=True,
            pick_up_offset=None,
            drop_offset=None,
        )
    )


def test_load_trash_bin(
    decoy: Decoy,
    mock_core: ProtocolCore,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should load a trash bin."""
    mock_trash = decoy.mock(cls=TrashBin)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(
            "blah", api_version, "OT-3 Standard"
        )
    ).then_return(DeckSlotName.SLOT_A1)
    decoy.when(
        mock_validation.ensure_and_convert_trash_bin_location(
            "blah", api_version, "OT-3 Standard"
        )
    ).then_return("my swanky trash bin")
    decoy.when(
        mock_core.load_trash_bin(DeckSlotName.SLOT_A1, "my swanky trash bin")
    ).then_return(mock_trash)

    result = subject.load_trash_bin("blah")

    assert result == mock_trash


def test_load_trash_bin_raises_for_staging_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should raise when a trash bin load is attempted in a staging slot."""
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(
            "bleh", api_version, "OT-3 Standard"
        )
    ).then_return(StagingSlotName.SLOT_A4)

    with pytest.raises(ValueError, match="Staging areas not permitted"):
        subject.load_trash_bin("bleh")


def test_load_waste_chute(
    decoy: Decoy,
    mock_core: ProtocolCore,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should load a waste chute."""
    mock_chute = decoy.mock(cls=WasteChute)
    decoy.when(mock_core.load_waste_chute()).then_return(mock_chute)
    result = subject.load_waste_chute()
    assert result == mock_chute


def test_load_module(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should load a module."""
    mock_module_core = decoy.mock(cls=TemperatureModuleCore)

    decoy.when(mock_validation.ensure_module_model("spline reticulator")).then_return(
        TemperatureModuleModel.TEMPERATURE_V1
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(DeckSlotName.SLOT_3)

    decoy.when(
        mock_core.load_module(
            model=TemperatureModuleModel.TEMPERATURE_V1,
            deck_slot=DeckSlotName.SLOT_3,
            configuration=None,
        )
    ).then_return(mock_module_core)

    decoy.when(mock_module_core.get_model()).then_return(
        TemperatureModuleModel.TEMPERATURE_V2
    )
    decoy.when(mock_module_core.get_serial_number()).then_return("cap'n crunch")
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_3)

    result = subject.load_module(module_name="spline reticulator", location=42)

    assert isinstance(result, ModuleContext)
    decoy.verify(mock_core_map.add(mock_module_core, result), times=1)


def test_load_module_default_location(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should load a module without specifying a location explicitly."""
    mock_module_core = decoy.mock(cls=TemperatureModuleCore)

    decoy.when(mock_validation.ensure_module_model("spline reticulator")).then_return(
        TemperatureModuleModel.TEMPERATURE_V1
    )

    decoy.when(
        mock_core.load_module(
            model=TemperatureModuleModel.TEMPERATURE_V1,
            deck_slot=None,
            configuration=None,
        )
    ).then_return(mock_module_core)

    decoy.when(mock_module_core.MODULE_TYPE).then_return(ModuleType.TEMPERATURE)
    decoy.when(mock_module_core.get_model()).then_return(
        TemperatureModuleModel.TEMPERATURE_V2
    )
    decoy.when(mock_module_core.get_serial_number()).then_return("cap'n crunch")
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_3)

    result = subject.load_module(module_name="spline reticulator", location=42)

    assert isinstance(result, ModuleContext)


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_load_module_with_configuration(subject: ProtocolContext) -> None:
    """It should raise an APIVersionError if the deprecated `configuration` argument is used."""
    with pytest.raises(UnsupportedAPIError):
        subject.load_module(
            module_name="spline reticulator",
            location=42,
            configuration="semi",
        )


@pytest.mark.parametrize("api_version", [APIVersion(2, 14)])
def test_load_module_with_mag_block_raises(subject: ProtocolContext) -> None:
    """It should raise an APIVersionError if loading a magnetic block."""
    with pytest.raises(UnsupportedAPIError):
        subject.load_module(
            module_name="magneticBlockV1",
            location=42,
            configuration="semi",
        )


def test_load_module_on_staging_slot_raises(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should raise when attempting to load a module onto a staging slot."""
    decoy.when(mock_validation.ensure_module_model("spline reticulator")).then_return(
        TemperatureModuleModel.TEMPERATURE_V1
    )
    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot(42, api_version, "OT-3 Standard")
    ).then_return(StagingSlotName.SLOT_B4)

    with pytest.raises(
        ValueError, match="Cannot load spline reticulator onto a staging slot."
    ):
        subject.load_module(module_name="spline reticulator", location=42)


def test_load_flex_stacker_on_staging_slot(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
) -> None:
    """It should load a module."""
    mock_module_core: FlexStackerCore = decoy.mock(cls=FlexStackerCore)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
    decoy.when(mock_validation.ensure_module_model("flexStackerModuleV1")).then_return(
        FlexStackerModuleModel.FLEX_STACKER_V1
    )
    decoy.when(
        mock_validation.ensure_and_convert_deck_slot("B4", api_version, "OT-3 Standard")
    ).then_return(StagingSlotName.SLOT_B4)
    decoy.when(
        mock_validation.convert_flex_stacker_load_slot(StagingSlotName.SLOT_B4)
    ).then_return(DeckSlotName.SLOT_B3)

    decoy.when(
        mock_core.load_module(
            model=FlexStackerModuleModel.FLEX_STACKER_V1,
            deck_slot=DeckSlotName.SLOT_B3,
            configuration=None,
        )
    ).then_return(mock_module_core)

    decoy.when(mock_module_core.get_model()).then_return(
        FlexStackerModuleModel.FLEX_STACKER_V1
    )
    decoy.when(mock_module_core.get_serial_number()).then_return("cap'n crunch")
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_B3)

    result = subject.load_module(module_name="flexStackerModuleV1", location="B4")

    assert isinstance(result, ModuleContext)
    decoy.verify(mock_core_map.add(mock_module_core, result), times=1)


@pytest.mark.parametrize(
    ("invert_load_order"),
    [
        True,
        False,
    ],
)
def test_load_flex_stacker_and_mag_block_combinations(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
    invert_load_order: bool,
) -> None:
    """It should load a Flex stacker and a Magnetic block in both acceptable orders."""
    mock_module_core_1: FlexStackerCore = decoy.mock(cls=FlexStackerCore)
    mock_module_core_2: MagneticBlockCore = decoy.mock(cls=MagneticBlockCore)

    decoy.when(mock_core.robot_type).then_return("OT-3 Standard")

    def _load_stacker() -> None:
        # Load a Flex Stacker at "B4" -> CutoutB3
        decoy.when(
            mock_validation.ensure_module_model("flexStackerModuleV1")
        ).then_return(FlexStackerModuleModel.FLEX_STACKER_V1)
        decoy.when(
            mock_validation.ensure_and_convert_deck_slot(
                "B4", api_version, "OT-3 Standard"
            )
        ).then_return(StagingSlotName.SLOT_B4)
        decoy.when(
            mock_validation.convert_flex_stacker_load_slot(StagingSlotName.SLOT_B4)
        ).then_return(DeckSlotName.SLOT_B3)

        decoy.when(
            mock_core.load_module(
                model=FlexStackerModuleModel.FLEX_STACKER_V1,
                deck_slot=DeckSlotName.SLOT_B3,
                configuration=None,
            )
        ).then_return(mock_module_core_1)

        decoy.when(mock_module_core_1.get_model()).then_return(
            FlexStackerModuleModel.FLEX_STACKER_V1
        )
        decoy.when(mock_module_core_1.get_serial_number()).then_return("cap'n crunch")
        decoy.when(mock_module_core_1.get_deck_slot()).then_return(DeckSlotName.SLOT_B3)

        result = subject.load_module(module_name="flexStackerModuleV1", location="B4")

        assert isinstance(result, ModuleContext)
        decoy.verify(mock_core_map.add(mock_module_core_1, result), times=1)

    def _load_mag() -> None:
        # Load a Mag Block on the B3 slot
        decoy.when(mock_validation.ensure_module_model("magneticBlockV1")).then_return(
            MagneticBlockModel.MAGNETIC_BLOCK_V1
        )
        decoy.when(
            mock_validation.ensure_and_convert_deck_slot(
                "B3", api_version, "OT-3 Standard"
            )
        ).then_return(DeckSlotName.SLOT_B3)
        decoy.when(
            mock_core.load_module(
                model=MagneticBlockModel.MAGNETIC_BLOCK_V1,
                deck_slot=DeckSlotName.SLOT_B3,
                configuration=None,
            )
        ).then_return(mock_module_core_2)
        result = subject.load_module(module_name="magneticBlockV1", location="B3")

        assert isinstance(result, ModuleContext)
        decoy.verify(mock_core_map.add(mock_module_core_2, result), times=1)

    if invert_load_order:
        _load_mag()
        _load_stacker()
    elif invert_load_order is False:
        _load_stacker()
        _load_mag()


@pytest.mark.parametrize(
    ("invert_load_order"),
    [
        True,
        False,
    ],
)
def test_load_flex_stacker_and_waste_chute_combinations(
    decoy: Decoy,
    mock_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    api_version: APIVersion,
    subject: ProtocolContext,
    invert_load_order: bool,
) -> None:
    """It should load a Flex stacker and a Waste Chute in both acceptable orders."""

    def _load_stacker() -> None:
        mock_module_core: FlexStackerCore = decoy.mock(cls=FlexStackerCore)
        decoy.when(mock_core.robot_type).then_return("OT-3 Standard")
        # Load a Flex Stacker at "B4" -> CutoutB3
        decoy.when(
            mock_validation.ensure_module_model("flexStackerModuleV1")
        ).then_return(FlexStackerModuleModel.FLEX_STACKER_V1)
        decoy.when(
            mock_validation.ensure_and_convert_deck_slot(
                "D4", api_version, "OT-3 Standard"
            )
        ).then_return(StagingSlotName.SLOT_D4)
        decoy.when(
            mock_validation.convert_flex_stacker_load_slot(StagingSlotName.SLOT_D4)
        ).then_return(DeckSlotName.SLOT_D3)

        decoy.when(
            mock_core.load_module(
                model=FlexStackerModuleModel.FLEX_STACKER_V1,
                deck_slot=DeckSlotName.SLOT_D3,
                configuration=None,
            )
        ).then_return(mock_module_core)

        decoy.when(mock_module_core.get_model()).then_return(
            FlexStackerModuleModel.FLEX_STACKER_V1
        )
        decoy.when(mock_module_core.get_serial_number()).then_return(
            "cooooookie crisp!"
        )
        decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_D3)

        result = subject.load_module(module_name="flexStackerModuleV1", location="D4")

        assert isinstance(result, ModuleContext)
        decoy.verify(mock_core_map.add(mock_module_core, result), times=1)

    def _load_chute() -> None:
        # Load a Waste Chute on the Flex Stacker's B3 slot
        mock_chute = decoy.mock(cls=WasteChute)
        decoy.when(mock_core.load_waste_chute()).then_return(mock_chute)
        result = subject.load_waste_chute()
        assert result == mock_chute

    if invert_load_order:
        _load_chute()
        _load_stacker()
    elif invert_load_order is False:
        _load_stacker()
        _load_chute()


def test_loaded_modules(
    decoy: Decoy,
    mock_core_map: LoadedCoreMap,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should return a list of all loaded modules."""
    module_core_4 = decoy.mock(cls=TemperatureModuleCore)
    module_core_6 = decoy.mock(cls=MagneticModuleCore)
    module_core_7 = decoy.mock(cls=MagneticBlockCore)

    module_4 = decoy.mock(cls=TemperatureModuleContext)
    module_6 = decoy.mock(cls=MagneticModuleContext)
    module_7 = decoy.mock(cls=MagneticBlockContext)

    decoy.when(mock_core.get_module_cores()).then_return(
        [module_core_4, module_core_6, module_core_7]
    )
    decoy.when(module_core_4.get_deck_slot()).then_return(DeckSlotName.SLOT_4)
    decoy.when(module_core_6.get_deck_slot()).then_return(DeckSlotName.SLOT_6)
    decoy.when(module_core_7.get_deck_slot()).then_return(DeckSlotName.SLOT_7)
    decoy.when(mock_core_map.get(module_core_4)).then_return(module_4)
    decoy.when(mock_core_map.get(module_core_6)).then_return(module_6)
    decoy.when(mock_core_map.get(module_core_7)).then_return(module_7)

    result = subject.loaded_modules

    assert result == {4: module_4, 6: module_6, 7: module_7}


def test_home(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should home all axes."""
    subject.home()
    decoy.verify(mock_core.home(), times=1)


def test_wait_for_tasks(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should wait for all tasks to complete."""
    task1 = decoy.mock(cls=Task)
    task2 = decoy.mock(cls=Task)
    decoy.when(task1._core).then_return(sentinel.task1_core)
    decoy.when(task2._core).then_return(sentinel.task2_core)
    subject.wait_for_tasks([task1, task2])
    decoy.verify(
        mock_core.wait_for_tasks([sentinel.task1_core, sentinel.task2_core]), times=1
    )


def test_create_timer(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should create a timer."""
    decoy.when(mock_core.create_timer(seconds=0.1)).then_return(sentinel.task1_core)
    result = subject.create_timer(seconds=0.1)
    assert result._core is sentinel.task1_core


def test_define_liquid(
    decoy: Decoy, mock_core: ProtocolCore, subject: ProtocolContext
) -> None:
    """It should add a liquid to the state."""
    expected_result = Liquid(
        _id="water-id",
        name="water",
        description="water desc",
        display_color="#1234",
    )

    decoy.when(
        mock_core.define_liquid(
            name="water", description="water desc", display_color="#1234"
        )
    ).then_return(expected_result)

    result = subject.define_liquid(
        name="water", description="water desc", display_color="#1234"
    )

    assert result == expected_result


@pytest.mark.parametrize(
    ("api_version", "expect_success"),
    [
        (APIVersion(2, 19), False),
        (APIVersion(2, 20), True),
    ],
)
def test_define_liquid_arg_defaulting(
    expect_success: bool,
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """Test API version dependent behavior for missing description and display_color."""
    success_result = Liquid(
        _id="water-id", name="water", description=None, display_color=None
    )
    decoy.when(
        mock_core.define_liquid(name="water", description=None, display_color=None)
    ).then_return(success_result)

    if expect_success:
        assert (
            subject.define_liquid(
                name="water"
                # description and display_color omitted.
            )
            == success_result
        )
    else:
        with pytest.raises(APIVersionError):
            subject.define_liquid(
                name="water"
                # description and display_color omitted.
            )


@pytest.mark.parametrize("robot_type", ["OT-2 Standard", "OT-3 Standard"])
def test_define_liquid_class(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
    robot_type: RobotType,
) -> None:
    """It should create the liquid class definition."""
    expected_liquid_class = LiquidClass(
        _name="volatile_100", _display_name="volatile 100%", _by_pipette_setting={}
    )
    decoy.when(mock_core.get_liquid_class("volatile_90", 123)).then_return(
        expected_liquid_class
    )
    decoy.when(mock_core.robot_type).then_return(robot_type)
    assert subject.get_liquid_class("volatile_90", 123) == expected_liquid_class


@pytest.mark.parametrize("robot_type", ["OT-2 Standard", "OT-3 Standard"])
def test_define_new_custom_liquid_class_from_dict(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
    robot_type: RobotType,
    minimal_transfer_properties_dict: Dict[str, Dict[str, TransferPropertiesDict]],
) -> None:
    """It should define a custom liquid class."""
    my_liquid_class = subject.define_liquid_class(
        name="my_liquid",
        properties=minimal_transfer_properties_dict,
        display_name="My liquid",
    )
    decoy.when(mock_core.robot_type).then_return(robot_type)
    my_liquid_class_props = my_liquid_class.get_for(
        "flex_1channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
    )
    assert my_liquid_class_props.aspirate.submerge.speed == 100
    assert (
        my_liquid_class_props.dispense.dispense_position.position_reference
        == PositionReference.WELL_BOTTOM
    )


@pytest.mark.parametrize("robot_type", ["OT-2 Standard", "OT-3 Standard"])
def test_customize_existing_liquid_class(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
    robot_type: RobotType,
    minimal_transfer_properties_dict: Dict[str, Dict[str, TransferPropertiesDict]],
    custom_pip_n_tip_transfer_properties_dict: Dict[
        str, Dict[str, TransferPropertiesDict]
    ],
) -> None:
    """It should create a new liquid class by modifying the existing liquid class."""
    existing_glycerol_class = LiquidClass.create(
        liquid_classes.load_definition("glycerol_50", 1)
    )
    assert (
        existing_glycerol_class.get_for(
            "flex_1channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ).aspirate.submerge.speed
        == 4
    )
    assert (
        existing_glycerol_class.get_for(
            "flex_8channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ).aspirate.submerge.speed
        == 4
    )

    my_liquid_class = subject.define_liquid_class(
        name="my_liquid",
        properties=minimal_transfer_properties_dict,
        base_liquid_class=existing_glycerol_class,
        display_name="My liquid",
    )
    decoy.when(mock_core.robot_type).then_return(robot_type)
    assert (
        my_liquid_class.get_for(
            "flex_1channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ).aspirate.submerge.speed
        == 100
    )
    assert (
        my_liquid_class.get_for(
            "flex_8channel_50", "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ).aspirate.submerge.speed
        == 4
    )

    # Test that new entries are created for pipettes and tipracks not present in the base liquid class
    lc_with_custom_pip_n_tip = subject.define_liquid_class(
        name="my_liquid_2",
        properties=custom_pip_n_tip_transfer_properties_dict,
        base_liquid_class=existing_glycerol_class,
        display_name="My liquid 2",
    )
    assert (
        lc_with_custom_pip_n_tip.get_for(
            "a_custom_pipette_type", "a_custom_tiprack_uri"
        ).aspirate.submerge.speed
        == 100
    )

    # Test that only specified tiprack's props are updated when there are
    # properties for multiple tipracks in the liquid class
    modified_custom_dict = custom_pip_n_tip_transfer_properties_dict.copy()
    modified_custom_dict["a_custom_pipette_type"] = {
        "some_new_tiprack": minimal_transfer_properties_dict["flex_1channel_50"][
            "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ]
    }
    lc_with_new_tip_entry = subject.define_liquid_class(
        name="my_liquid_3",
        properties=modified_custom_dict,
        base_liquid_class=lc_with_custom_pip_n_tip,
        display_name="My liquid 3",
    )
    assert (
        lc_with_new_tip_entry.get_for(
            "a_custom_pipette_type", "a_custom_tiprack_uri"
        ).aspirate.submerge.speed
        == lc_with_new_tip_entry.get_for(
            "a_custom_pipette_type", "some_new_tiprack"
        ).aspirate.submerge.speed
        == 100
    )


def test_capture_image(
    decoy: Decoy,
    mock_core: ProtocolCore,
    subject: ProtocolContext,
) -> None:
    """It should handle the core execution for capture image using the provided parameters or defaults."""
    subject.capture_image(
        home_before=True,
        filename="coolpic",
        resolution=(1920, 1080),
        zoom=2,
        contrast=50,
        saturation=52,
    )

    decoy.verify(mock_core.home())
    decoy.verify(
        mock_core.capture_image(
            filename="coolpic",
            resolution=(1920, 1080),
            zoom=2,
            contrast=50,
            brightness=None,
            saturation=52,
        )
    )


def test_bundled_data(
    mock_core_map: LoadedCoreMap, mock_deck: Deck, mock_core: ProtocolCore
) -> None:
    """It should return bundled data."""
    subject = ProtocolContext(
        api_version=MAX_SUPPORTED_VERSION,
        core=mock_core,
        core_map=mock_core_map,
        deck=mock_deck,
        bundled_data={"foo": b"ar"},
    )

    assert subject.bundled_data == {"foo": b"ar"}



import asyncio
from typing import AsyncGenerator

import pytest

from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control import ExecutionManager, modules
from opentrons.hardware_control.modules.magdeck import MagDeck
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
)


@pytest.fixture
def usb_port() -> USBPort:
    """Token USB port."""
    return USBPort(
        name="",
        port_number=0,
        device_path="/dev/ot_module_sim_magdeck0",
    )


@pytest.fixture
async def subject(
    usb_port: USBPort,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
) -> AsyncGenerator[modules.AbstractModule, None]:
    """Test subject."""
    mag = await modules.build(
        port="/dev/ot_module_sim_magdeck0",
        usb_port=usb_port,
        type=modules.ModuleType["MAGNETIC"],
        simulating=True,
        hw_control_loop=asyncio.get_running_loop(),
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    try:
        yield mag
    finally:
        await mag.cleanup()


async def test_sim_initialization(subject: modules.MagDeck) -> None:
    """It should initialize to an AbstractModule."""
    assert isinstance(subject, modules.AbstractModule)


async def test_sim_data(subject: modules.MagDeck) -> None:
    """It should forward simulated data."""
    assert subject.status == "disengaged"
    assert subject.device_info["serial"] == "dummySerialMD"
    # return v1 when sim_model is not passed
    assert subject.device_info["model"] == "mag_deck_v1.1"
    assert subject.device_info["version"] == "dummyVersionMD"
    assert subject.live_data["status"] == subject.status
    assert "data" in subject.live_data


async def test_sim_state_update(subject: modules.MagDeck) -> None:
    """It should update simulated state."""
    assert isinstance(subject, MagDeck)
    await subject.calibrate()
    assert subject.status == "disengaged"
    await subject.engage(2)
    assert subject.status == "engaged"
    await subject.deactivate()
    assert subject.status == "disengaged"


async def test_revision_model_parsing(subject: modules.MagDeck) -> None:
    """It should parse its own revision."""
    assert isinstance(subject, MagDeck)
    subject._device_info["model"] = "mag_deck_v1.1"
    assert subject.model() == "magneticModuleV1"
    subject._device_info["model"] = "mag_deck_v20"
    assert subject.model() == "magneticModuleV2"
    del subject._device_info["model"]
    assert subject.model() == "magneticModuleV1"



import asyncio
from contextlib import nullcontext as does_not_raise
from typing import Any, AsyncGenerator, ContextManager

import mock
import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    FlexStackerHopperLabwareError,
    FlexStackerShuttleLabwareError,
    FlexStackerShuttleNotEmptyError,
)

from opentrons.drivers.flex_stacker.simulator import SimulatingDriver
from opentrons.drivers.flex_stacker.types import (
    Direction,
    HardwareRevision,
    LEDColor,
    LEDPattern,
    LimitSwitchStatus,
    PlatformStatus,
    StackerAxis,
    StackerInfo,
    TOFSensor,
    TOFSensorMode,
    TOFSensorState,
    TOFSensorStatus,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.hardware_control import ExecutionManager, modules
from opentrons.hardware_control.modules.flex_stacker import (
    HOME_OFFSET_MD,
    HOME_OFFSET_SM,
    LATCH_CLEARANCE,
    MAX_TRAVEL,
    PLATFORM_OFFSET,
    SIMULATING_POLL_PERIOD,
    STACKER_MOTION_CONFIG,
    FlexStackerReader,
)
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    ModuleErrorCallback,
    PlatformState,
)
from opentrons.hardware_control.poller import Poller
from opentrons.hardware_control.types import StatusBarState, StatusBarUpdateEvent


@pytest.fixture
def usb_port() -> USBPort:
    """Token USB port."""
    return USBPort(
        name="",
        port_number=0,
        device_path="/dev/ot_module_sim_flexstacker0",
    )


@pytest.fixture
def mock_driver(decoy: Decoy) -> SimulatingDriver:
    """Mocked simulating driver."""
    return decoy.mock(cls=SimulatingDriver)


@pytest.fixture
async def subject(
    usb_port: USBPort,
    mock_driver: SimulatingDriver,
    mock_execution_manager: ExecutionManager,
    module_error_callback: ModuleErrorCallback,
    module_disconnected_callback: ModuleDisconnectedCallback,
    decoy: Decoy,
) -> AsyncGenerator[modules.FlexStacker, None]:
    """Test subject with mocked driver."""
    reader = FlexStackerReader(driver=mock_driver)
    poller = Poller(reader=reader, interval=SIMULATING_POLL_PERIOD)
    stacker = modules.FlexStacker(
        port="/dev/ot_module_sim_flexstacker0",
        usb_port=usb_port,
        driver=mock_driver,
        reader=reader,
        poller=poller,
        device_info={
            "serial": "dummySerialFS",
            "model": "a1",
            "version": "stacker-fw",
        },
        hw_control_loop=asyncio.get_running_loop(),
        execution_manager=mock_execution_manager,
        error_callback=module_error_callback,
        disconnected_callback=module_disconnected_callback,
    )
    decoy.when(await mock_driver.get_device_info()).then_return(
        StackerInfo(fw="stacker-fw", hw=HardwareRevision.EVT, sn="dummySerialFS")
    )
    decoy.when(await mock_driver.get_tof_sensor_status(TOFSensor.X)).then_return(
        TOFSensorStatus(
            TOFSensor.X, TOFSensorState.IDLE, TOFSensorMode.MEASURE, ok=True
        )
    )
    decoy.when(await mock_driver.get_tof_sensor_status(TOFSensor.Z)).then_return(
        TOFSensorStatus(
            TOFSensor.Z, TOFSensorState.IDLE, TOFSensorMode.MEASURE, ok=True
        )
    )
    decoy.when(await mock_driver.get_platform_status()).then_return(
        PlatformStatus(False, False)
    )
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(
        LimitSwitchStatus(False, True, False, False, False)
    )

    await poller.start()
    try:
        yield stacker
    finally:
        await stacker.cleanup()


async def test_sim_state(subject: modules.FlexStacker) -> None:
    """It should forward state."""
    status = subject.device_info
    assert status["serial"] == "dummySerialFS"
    assert status["model"] == "a1"
    assert status["version"] == "stacker-fw"


async def test_set_run_hold_current(
    subject: modules.FlexStacker, mock_driver: SimulatingDriver, decoy: Decoy
) -> None:
    """It should set currents."""
    # Test move_axis

    # run and hold current are 0 by default
    assert subject._reader.motion_params[StackerAxis.X].run_current == 0
    assert subject._reader.motion_params[StackerAxis.X].hold_current == 0
    default = STACKER_MOTION_CONFIG[StackerAxis.X]["move"]

    # Call the move_axis function with default current
    await subject.move_axis(StackerAxis.X, Direction.EXTEND, 44)
    # set_run_current should be called and run_current recorded
    decoy.verify(await mock_driver.set_run_current(StackerAxis.X, default.run_current))
    decoy.verify(
        await mock_driver.set_ihold_current(StackerAxis.X, default.hold_current)
    )
    motion_params = subject._reader.motion_params[StackerAxis.X]
    assert motion_params.run_current == default.run_current
    assert motion_params.hold_current == default.hold_current
    decoy.reset()

    decoy.when(await mock_driver.get_platform_status()).then_return(
        PlatformStatus(False, True)
    )
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(
        LimitSwitchStatus(True, True, False, False, False)
    )

    # Make sure set_run_current and set_ihold_current are not called again
    await subject.move_axis(StackerAxis.X, Direction.EXTEND, 44)
    decoy.verify(
        await mock_driver.set_run_current(matchers.Anything(), matchers.Anything()),
        times=0,
    )
    decoy.verify(
        await mock_driver.set_ihold_current(matchers.Anything(), matchers.Anything()),
        times=0,
    )
    motion_params = subject._reader.motion_params[StackerAxis.X]
    assert motion_params.run_current == default.run_current
    assert motion_params.hold_current == default.hold_current

    # Test home_axis
    decoy.reset()
    # Reset the run/hold current recorded
    default = STACKER_MOTION_CONFIG[StackerAxis.X]["home"]
    subject._reader.motion_params[StackerAxis.X].run_current = 0
    subject._reader.motion_params[StackerAxis.X].hold_current = 0
    decoy.when(await mock_driver.get_platform_status()).then_return(
        PlatformStatus(False, True)
    )
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(
        LimitSwitchStatus(True, True, False, False, False)
    )

    # Call the home_axis function with default current
    await subject.home_axis(StackerAxis.X, Direction.EXTEND)
    decoy.verify(await mock_driver.set_run_current(StackerAxis.X, default.run_current))
    decoy.verify(
        await mock_driver.set_ihold_current(StackerAxis.X, default.hold_current)
    )
    motion_params = subject._reader.motion_params[StackerAxis.X]
    assert motion_params.run_current == default.run_current
    assert motion_params.hold_current == default.hold_current
    decoy.reset()
    decoy.when(await mock_driver.get_platform_status()).then_return(
        PlatformStatus(False, True)
    )
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(
        LimitSwitchStatus(True, True, False, False, False)
    )

    # Make sure set_run_current and set_ihold_current are not called again
    await subject.home_axis(StackerAxis.X, Direction.EXTEND, 44)
    decoy.verify(
        await mock_driver.set_run_current(matchers.Anything(), matchers.Anything()),
        times=0,
    )
    decoy.verify(
        await mock_driver.set_ihold_current(matchers.Anything(), matchers.Anything()),
        times=0,
    )

    # The recorded run/hold current should stay the same
    motion_params = subject._reader.motion_params[StackerAxis.X]
    assert motion_params.run_current == default.run_current
    assert motion_params.hold_current == default.hold_current


PLATFORM_STATUS_UNKNOWN = PlatformStatus(False, False)
PLATFORM_STATUS_EXTENDED = PlatformStatus(True, False)
PLATFORM_STATUS_RETRACTED = PlatformStatus(False, True)

X_UNKNOWN = LimitSwitchStatus(False, False, False, False, False)
X_EXTENDED = LimitSwitchStatus(True, False, False, False, False)
X_RETRACTED = LimitSwitchStatus(False, True, False, False, False)


@pytest.mark.parametrize("x_status", [X_EXTENDED, X_RETRACTED, X_UNKNOWN])
@pytest.mark.parametrize(
    "platform_status,expected",
    [
        (PLATFORM_STATUS_RETRACTED, PlatformState.RETRACTED),
        (PLATFORM_STATUS_EXTENDED, PlatformState.EXTENDED),
    ],
)
async def test_platform_state(
    subject: modules.FlexStacker,
    mock_driver: SimulatingDriver,
    x_status: LimitSwitchStatus,
    platform_status: PlatformStatus,
    expected: PlatformState,
    decoy: Decoy,
) -> None:
    """Test that the platform state is correctly determined."""
    decoy.when(await mock_driver.get_platform_status()).then_return(platform_status)
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(x_status)

    # update the cached value
    await subject._reader.get_limit_switch_status()
    await subject._reader.get_platform_sensor_state()
    assert subject.platform_state == expected


@pytest.mark.parametrize(
    "x_status,expected",
    [
        (X_EXTENDED, PlatformState.MISSING),
        (X_RETRACTED, PlatformState.MISSING),
        (X_UNKNOWN, PlatformState.UNKNOWN),
    ],
)
async def test_platform_state_unknown(
    subject: modules.FlexStacker,
    mock_driver: SimulatingDriver,
    x_status: LimitSwitchStatus,
    expected: PlatformState,
    decoy: Decoy,
) -> None:
    """Test that the platform state is correctly determined."""

    decoy.when(await mock_driver.get_platform_status()).then_return(
        PLATFORM_STATUS_UNKNOWN
    )
    decoy.when(await mock_driver.get_limit_switches_status()).then_return(x_status)

    # update the value
    await subject._reader.get_limit_switch_status()
    await subject._reader.get_platform_sensor_state()
    assert subject.platform_state == expected


@pytest.mark.parametrize(
    ("should_identify", "hopper_door", "event", "result_params"),
    [
        (  # running
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.RUNNING, enabled=True),
            (0.5, LEDColor.GREEN, LEDPattern.STATIC, None),
        ),
        (  # paused - door open
            False,
            False,
            StatusBarUpdateEvent(state=StatusBarState.PAUSED, enabled=True),
            (0.5, LEDColor.BLUE, LEDPattern.PULSE, 2000),
        ),
        (  # paused - should identify
            True,
            True,
            StatusBarUpdateEvent(state=StatusBarState.PAUSED, enabled=True),
            (0.5, LEDColor.BLUE, LEDPattern.PULSE, 2000),
        ),
        (  # paused - door closed not identified
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.PAUSED, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # idle - door open
            False,
            False,
            StatusBarUpdateEvent(state=StatusBarState.IDLE, enabled=True),
            (0.5, LEDColor.BLUE, LEDPattern.PULSE, 2000),
        ),
        (  # idle - door closed
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.IDLE, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # hardware error - identified
            True,
            True,
            StatusBarUpdateEvent(state=StatusBarState.HARDWARE_ERROR, enabled=True),
            (0.5, LEDColor.RED, LEDPattern.FLASH, 300),
        ),
        (  # hardware error - not identified
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.HARDWARE_ERROR, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # software error
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.SOFTWARE_ERROR, enabled=True),
            (0.5, LEDColor.YELLOW, LEDPattern.STATIC, None),
        ),
        (  # error recovery - door open
            False,
            False,
            StatusBarUpdateEvent(state=StatusBarState.ERROR_RECOVERY, enabled=True),
            (0.5, LEDColor.BLUE, LEDPattern.PULSE, 2000),
        ),
        (  # error recovery - should identify
            True,
            True,
            StatusBarUpdateEvent(state=StatusBarState.ERROR_RECOVERY, enabled=True),
            (0.5, LEDColor.YELLOW, LEDPattern.PULSE, 2000),
        ),
        (  # error recovery - door closed
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.ERROR_RECOVERY, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.STATIC, None),
        ),
        (  # run complete
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.RUN_COMPLETED, enabled=True),
            (0.5, LEDColor.GREEN, LEDPattern.PULSE, None),
        ),
        (  # updating
            False,
            True,
            StatusBarUpdateEvent(state=StatusBarState.UPDATING, enabled=True),
            (0.5, LEDColor.WHITE, LEDPattern.PULSE, None),
        ),
    ],
)
async def test_stacker_status_bar_event_handler(
    subject: modules.FlexStacker,
    mock_driver: SimulatingDriver,
    should_identify: bool,
    hopper_door: bool,
    event: StatusBarUpdateEvent,
    result_params: tuple[float, LEDColor, LEDPattern, int | None],
    decoy: Decoy,
) -> None:
    """It should handle LED lights."""
    decoy.when(await mock_driver.get_hopper_door_closed()).then_return(hopper_door)
    subject.set_stacker_identify(should_identify)
    await subject._reader.get_door_closed()
    await subject._handle_status_bar_event(event)
    decoy.verify(
        await mock_driver.set_led(
            result_params[0],
            color=result_params[1],
            pattern=result_params[2],
            duration=result_params[3],
            reps=None,
        )
    )


@pytest.mark.parametrize(
    ("labware_height"),
    [(16), (100)],
)
async def test_store_labware_motion_sequence(
    subject: modules.FlexStacker, labware_height: float
) -> None:
    """It should store labware with sensing on."""
    with (
        mock.patch.object(
            subject, "_prepare_for_action", mock.AsyncMock()
        ) as _prepare_for_action,
        mock.patch.object(
            subject, "_move_and_home_axis", mock.AsyncMock()
        ) as _move_and_home_axis,
        mock.patch.object(
            subject, "verify_shuttle_labware_presence", mock.AsyncMock()
        ) as verify_shuttle_labware_presence,
        mock.patch.object(subject, "move_axis", mock.AsyncMock()) as move_axis,
        mock.patch.object(subject, "home_axis", mock.AsyncMock()) as home_axis,
        mock.patch.object(subject, "open_latch", mock.AsyncMock()) as open_latch,
        mock.patch.object(subject, "close_latch", mock.AsyncMock()) as close_latch,
    ):
        # Test valid labware height
        await subject.store_labware(
            labware_height=labware_height,
            enforce_shuttle_lw_sensing=True,
        )

        # We need to verify the move sequence
        _prepare_for_action.assert_called()
        _move_and_home_axis.assert_any_call(
            StackerAxis.X, Direction.RETRACT, HOME_OFFSET_MD
        )
        verify_shuttle_labware_presence.assert_any_call(Direction.RETRACT, True)

        # Assertions for offset calculation and move_axis
        latch_clear_distance = labware_height + PLATFORM_OFFSET - LATCH_CLEARANCE
        distance = MAX_TRAVEL[StackerAxis.Z] - latch_clear_distance
        move_axis.assert_any_call(StackerAxis.Z, Direction.EXTEND, distance)

        # Verify labware transfer
        open_latch.assert_called_once()
        z_distance = latch_clear_distance - HOME_OFFSET_SM
        z_speed = STACKER_MOTION_CONFIG[StackerAxis.Z]["move"].move_params.max_speed / 2
        move_axis.assert_any_call(StackerAxis.Z, Direction.EXTEND, z_distance, z_speed)
        home_axis.assert_any_call(StackerAxis.Z, Direction.EXTEND, z_speed)
        close_latch.assert_called_once()

        # Now the z can be moved down and verify no labware is detected
        _move_and_home_axis.assert_any_call(
            StackerAxis.Z, Direction.RETRACT, HOME_OFFSET_MD
        )
        verify_shuttle_labware_presence.assert_any_call(Direction.RETRACT, False)

        # Then finally the x is moved to the gripper position
        _move_and_home_axis.assert_any_call(
            StackerAxis.X, Direction.EXTEND, HOME_OFFSET_MD
        )


@pytest.mark.parametrize(
    ("labware_height"),
    [(16), (100)],
)
async def test_dispense_labware_motion_sequence(
    subject: modules.FlexStacker,
    labware_height: float,
) -> None:
    """
    Test successful dispense labware with labware sensing enforced.
    """
    with (
        mock.patch.object(
            subject, "_prepare_for_action", mock.AsyncMock()
        ) as _prepare_for_action,
        mock.patch.object(
            subject, "_move_and_home_axis", mock.AsyncMock()
        ) as _move_and_home_axis,
        mock.patch.object(
            subject, "labware_detected", mock.AsyncMock()
        ) as labware_detected,
        mock.patch.object(
            subject,
            "verify_shuttle_labware_presence",
            autospec=True,
            side_effect=subject.verify_shuttle_labware_presence,
        ) as verify_shuttle_labware_presence,
        mock.patch.object(
            subject, "verify_hopper_labware_presence", autospec=True
        ) as verify_hopper_labware_presence,
        mock.patch.object(subject, "move_axis", mock.AsyncMock()) as move_axis,
        mock.patch.object(subject, "home_axis", mock.AsyncMock()) as home_axis,
        mock.patch.object(subject, "open_latch", mock.AsyncMock()) as open_latch,
        mock.patch.object(subject, "close_latch", mock.AsyncMock()) as close_latch,
    ):
        labware_detected.side_effect = [True, False, True]
        # Test valid labware height
        await subject.dispense_labware(
            labware_height=labware_height,
        )

        # We need to verify the move sequence
        # verify_hopper_labware_presence.assert_called_once_with(Direction.EXTEND, True)
        verify_hopper_labware_presence.assert_not_called()
        _prepare_for_action.assert_called()

        _move_and_home_axis.assert_any_call(
            StackerAxis.X, Direction.RETRACT, HOME_OFFSET_MD
        )
        # Verify labware presence
        verify_shuttle_labware_presence.assert_any_call(Direction.RETRACT, False)
        _move_and_home_axis.assert_any_call(
            StackerAxis.Z, Direction.EXTEND, HOME_OFFSET_SM
        )

        # Verify labware transfer
        open_latch.assert_called_once()
        latch_clear_distance = labware_height + PLATFORM_OFFSET - LATCH_CLEARANCE
        move_axis.assert_any_call(
            StackerAxis.Z, Direction.RETRACT, latch_clear_distance
        )
        close_latch.assert_called_once()

        # Assertions for offset calculation and move_axis/home_axis
        z_distance = MAX_TRAVEL[StackerAxis.Z] - latch_clear_distance - HOME_OFFSET_SM
        move_axis.assert_any_call(StackerAxis.Z, Direction.RETRACT, z_distance)
        home_axis.assert_any_call(StackerAxis.Z, Direction.RETRACT)

        # Verify labware presence
        verify_shuttle_labware_presence.assert_any_call(Direction.RETRACT, True)

        # Then finally the x is moved to the gripper position
        _move_and_home_axis.assert_any_call(
            StackerAxis.X, Direction.EXTEND, HOME_OFFSET_MD
        )

        # Make sure labware presense check on the X retract was called twice
        labware_detected.assert_has_calls(
            [
                mock.call(StackerAxis.Z, Direction.EXTEND),
                mock.call(StackerAxis.X, Direction.RETRACT),
                mock.call(StackerAxis.X, Direction.RETRACT),
            ]
        )
        assert verify_shuttle_labware_presence.call_count == 2


@pytest.mark.parametrize("hopper_lw_detected", [True, False])
@pytest.mark.parametrize("shuttle_lw_detected_before", [True, False])
@pytest.mark.parametrize("shuttle_lw_detected_after", [True, False])
async def test_dispense_labware_error_handling(
    subject: modules.FlexStacker,
    hopper_lw_detected: bool,
    shuttle_lw_detected_before: bool,
    shuttle_lw_detected_after: bool,
) -> None:
    """
    Test different error handling scenarios when dispensing labware.
    """
    expected_raise: ContextManager[Any]
    if shuttle_lw_detected_before is True:
        # If the shuttle is occupied, it should always raise an error
        expected_raise = pytest.raises(FlexStackerShuttleNotEmptyError)
    elif shuttle_lw_detected_after is True:
        # if the shuttle has labware after dispensing, no need to raise any error
        expected_raise = does_not_raise()
    elif hopper_lw_detected is False:
        # if the shuttle has no labware after dispensing, and the hopper has no labware,
        # it should raise a FlexStackerHopperLabwareError
        expected_raise = pytest.raises(FlexStackerHopperLabwareError)
    else:
        # if the shuttle has no labware after dispensing, but the hopper has labware,
        # it should raise a FlexStackerShuttleLabwareError letting user know
        # the labware latch is properly jammed
        expected_raise = pytest.raises(FlexStackerShuttleLabwareError)

    with (
        mock.patch.object(
            subject,
            "labware_detected",
            side_effect=[
                hopper_lw_detected,
                shuttle_lw_detected_before,
                shuttle_lw_detected_after,
            ],
            autospec=True,
        ) as labware_detected,
    ):
        with expected_raise:
            # Test valid labware height
            await subject.dispense_labware(
                labware_height=100.0,
            )

        expected_calls = [
            mock.call(StackerAxis.Z, Direction.EXTEND),
            mock.call(StackerAxis.X, Direction.RETRACT),
            mock.call(StackerAxis.X, Direction.RETRACT),
        ]
        if shuttle_lw_detected_before:
            assert labware_detected.call_count == 2
            labware_detected.assert_has_calls(expected_calls[:2])
        else:
            assert labware_detected.call_count == 3
            labware_detected.assert_has_calls(expected_calls)


@pytest.mark.parametrize(
    ("labware_height"),
    [(0), (-10), (200)],
)
async def test_invalid_labware_height(
    subject: modules.FlexStacker, labware_height: float
) -> None:
    """Raise a ValueError if the labware_height is invalid"""
    with mock.patch.object(
        subject, "_prepare_for_action", mock.AsyncMock()
    ) as _prepare_for_action:
        # Test invalid labware height
        with pytest.raises(ValueError):
            await subject.store_labware(
                labware_height=labware_height,
                enforce_shuttle_lw_sensing=True,
            )
        _prepare_for_action.assert_not_called()


async def test_error_callback(
    subject: modules.FlexStacker,
    mock_driver: SimulatingDriver,
    module_error_callback: ModuleErrorCallback,
    decoy: Decoy,
) -> None:
    """It should forward an error from the poller to the callback."""
    # There are no asynchronous errors from the stacker (the door is handled
    # differently) so just inject a random exception.
    exc = Exception("Oh no!")
    decoy.when(mock_driver.get_hopper_door_closed()).then_raise(exc)
    with pytest.raises(Exception, match="Oh no!"):
        await subject._poller.wait_next_poll()
    decoy.verify(
        module_error_callback(
            exc,
            "flexStackerModuleV1",
            "/dev/ot_module_sim_flexstacker0",
            "dummySerialFS",
        )
    )



"""Tests for the AddressableAreaStore+AddressableAreaState+AddressableAreaView trifecta.

The trifecta is tested here as a single unit, treating AddressableAreaState as a private
implementation detail.
"""

from opentrons_shared_data.deck.types import DeckDefinitionV5

from opentrons.protocol_engine.actions.actions import SetDeckConfigurationAction
from opentrons.protocol_engine.state.addressable_areas import (
    AddressableAreaStore,
    AddressableAreaView,
)
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.types import DeckType


def test_deck_configuration_setting(
    ot3_standard_deck_def: DeckDefinitionV5,
) -> None:
    """You should be able to set the deck configuration with a SetDeckConfigurationAction."""
    subject = AddressableAreaStore(
        deck_configuration=[("cutoutA3", "trashBinAdapter", None)],
        config=Config(
            use_simulated_deck_config=False,
            # Doesn't matter:
            robot_type="OT-3 Standard",
            deck_type=DeckType.OT3_STANDARD,
        ),
        deck_definition=ot3_standard_deck_def,
        robot_definition={
            "displayName": "OT-3",
            "robotType": "OT-3 Standard",
            "models": ["OT-3 Standard"],
            "extents": [477.2, 493.8, 0.0],
            "paddingOffsets": {
                "rear": -177.42,
                "front": 51.8,
                "leftSide": 31.88,
                "rightSide": -80.32,
            },
            "mountOffsets": {
                "left": [-13.5, -60.5, 255.675],
                "right": [40.5, -60.5, 255.675],
                "gripper": [84.55, -12.75, 93.85],
            },
        },
    )
    subject_view = AddressableAreaView(subject.state)

    assert subject_view.get_all_cutout_fixtures() == ["trashBinAdapter"]

    # The new deck configuration should overwrite the old one.
    subject.handle_action(
        SetDeckConfigurationAction(
            deck_configuration=[
                ("cutoutA3", "trashBinAdapter", None),
                ("cutoutB3", "singleRightSlot", None),
            ],
        )
    )
    assert sorted(subject_view.get_all_cutout_fixtures() or []) == [
        "singleRightSlot",
        "trashBinAdapter",
    ]

    # Deck configurations of None should no-op.
    subject.handle_action(
        SetDeckConfigurationAction(
            deck_configuration=None,
        )
    )
    assert sorted(subject_view.get_all_cutout_fixtures() or []) == [
        "singleRightSlot",
        "trashBinAdapter",
    ]



"""Tests for the opentrons.protocols.geometry.Deck interface."""

from typing import Any

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.labware.types import LabwareUri

from opentrons.motion_planning import deck_conflict
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.module_geometry import (
    HeaterShakerGeometry,
    ModuleGeometry,
    ThermocyclerGeometry,
)
from opentrons.protocol_api.labware import Labware
from opentrons.protocols.api_support.deck_type import (
    SHORT_TRASH_DECK,
    STANDARD_OT2_DECK,
)
from opentrons.types import DeckSlotName


@pytest.fixture(autouse=True)
def use_mock_deck_conflict_check(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out the deck conflict checker."""
    mock_check = decoy.mock(func=deck_conflict.check)
    monkeypatch.setattr(deck_conflict, "check", mock_check)


# We expect the subject to call the deck conflict checker with a fixed trash argument
# that matches this.
EXPECTED_FIXED_TRASH = deck_conflict.Labware(
    uri=matchers.Anything(),
    highest_z=matchers.Anything(),
    is_fixed_trash=True,
    name_for_errors=matchers.Anything(),
)


# This Deck class is only used by Python Protocol API versions earlier than 2.13,
# which only support the OT-2, not the OT-3.
@pytest.fixture(params=[STANDARD_OT2_DECK, SHORT_TRASH_DECK])
def subject(request: Any) -> Deck:
    """Get a Deck test subject."""
    return Deck(deck_type=request.param)


@pytest.mark.parametrize("slot_number", range(1, 12))
def test_slot_names(decoy: Decoy, slot_number: int, subject: Deck) -> None:
    """Ensure item get/set/delete works with both strs and ints."""
    deck_item = decoy.mock(cls=LegacyLabwareCore)
    decoy.when(deck_item.get_quirks()).then_return([])
    decoy.when(deck_item.highest_z).then_return(42)

    slot_name = str(slot_number)

    subject[slot_number] = deck_item
    assert subject[slot_number] is deck_item
    assert subject[slot_name] is deck_item

    del subject[slot_number]
    assert subject[slot_number] is None
    assert subject[slot_name] is None

    subject[slot_name] = deck_item
    assert subject[slot_number] is deck_item
    assert subject[slot_name] is deck_item

    del subject[slot_name]
    assert subject[slot_number] is None
    assert subject[slot_name] is None


@pytest.mark.parametrize("bad_slot", [0, 13, 1.0, "0", "13", "1.0", "hello world"])
def test_invalid_slot_names(decoy: Decoy, bad_slot: Any, subject: Deck) -> None:
    """It should reject things that aren't slot names."""
    with pytest.raises(ValueError, match="Unknown slot"):
        subject[bad_slot]

    with pytest.raises(ValueError, match="Unknown slot"):
        subject[bad_slot] = decoy.mock(cls=LegacyLabwareCore)


def test_highest_z(decoy: Decoy, subject: Deck) -> None:
    """It should return the highest Z point of all objects on the deck."""
    item_10 = decoy.mock(cls=LegacyLabwareCore)
    decoy.when(item_10.get_quirks()).then_return([])
    decoy.when(item_10.highest_z).then_return(10)

    item_100 = decoy.mock(cls=LegacyLabwareCore)
    decoy.when(item_100.get_quirks()).then_return([])
    decoy.when(item_100.highest_z).then_return(100)

    del subject[12]
    assert subject.highest_z == 0

    subject[1] = item_10
    assert subject.highest_z == 10

    subject[2] = item_100
    assert subject.highest_z == 100

    decoy.when(item_100.highest_z).then_return(50)
    subject.recalculate_high_z()
    assert subject.highest_z == 50

    del subject[1]
    assert subject.highest_z == 50

    del subject[2]
    assert subject.highest_z == 0


def test_fixed_trash_conflict_checking(decoy: Decoy) -> None:
    """It should correctly call the deck conflict checker with fixed trash labware."""
    subject = Deck(deck_type=STANDARD_OT2_DECK)
    decoy.verify(
        deck_conflict.check(
            existing_items={},
            new_location=DeckSlotName.FIXED_TRASH,
            new_item=EXPECTED_FIXED_TRASH,
            robot_type="OT-2 Standard",
        )
    )

    # Load another trash as an opentrons.protocol_api.Labware and test that the subject
    # correctly calls the deck conflict checker with that, too.
    # This intended to mimic what happens when we run a JSON protocol of v5 or below.
    # A trash labware is added to the subject that overrides the one that it loaded
    # upon construction.
    another_trash = decoy.mock(cls=Labware)
    decoy.when(another_trash.highest_z).then_return(42)
    decoy.when(another_trash.load_name).then_return("trash_load_name")
    decoy.when(another_trash.quirks).then_return(["fixedTrash"])
    decoy.when(another_trash.uri).then_return("test/trash/1")
    subject["12"] = another_trash
    decoy.verify(
        deck_conflict.check(
            existing_items={DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH},
            new_location=DeckSlotName.FIXED_TRASH,
            new_item=deck_conflict.Labware(
                uri=LabwareUri("test/trash/1"),
                highest_z=42,
                is_fixed_trash=True,
                name_for_errors="trash_load_name",
            ),
            robot_type="OT-2 Standard",
        )
    )


def test_labware_conflict_checking(
    decoy: Decoy,
    subject: Deck,
) -> None:
    """It should correctly call the deck conflict checker when given a labware."""
    # When given an internal LegacyLabwareCore object:

    input_legacy_labware_core = decoy.mock(cls=LegacyLabwareCore)
    decoy.when(input_legacy_labware_core.highest_z).then_return(42)
    decoy.when(input_legacy_labware_core.load_name).then_return(
        "legacy_labware_core_load_name"
    )
    decoy.when(input_legacy_labware_core.get_quirks()).then_return([])
    decoy.when(input_legacy_labware_core.get_uri()).then_return(
        "legacy_labware_core_uri"
    )

    subject["4"] = input_legacy_labware_core

    decoy.verify(
        deck_conflict.check(
            existing_items={
                DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH,
            },
            new_location=DeckSlotName.SLOT_4,
            new_item=deck_conflict.Labware(
                uri=LabwareUri("legacy_labware_core_uri"),
                highest_z=42,
                is_fixed_trash=False,
                name_for_errors="legacy_labware_core_load_name",
            ),
            robot_type="OT-2 Standard",
        ),
        times=1,
    )

    del subject["4"]

    # When given a public-facing Labware object:

    input_labware = decoy.mock(cls=Labware)
    decoy.when(input_labware.highest_z).then_return(42)
    decoy.when(input_labware.load_name).then_return("labware_load_name")
    decoy.when(input_labware.quirks).then_return([])
    decoy.when(input_labware.uri).then_return("labware_uri")

    subject["4"] = input_labware

    decoy.verify(
        deck_conflict.check(
            existing_items={
                DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH,
            },
            new_location=DeckSlotName.SLOT_4,
            new_item=deck_conflict.Labware(
                uri=LabwareUri("labware_uri"),
                highest_z=42,
                is_fixed_trash=False,
                name_for_errors="labware_load_name",
            ),
            robot_type="OT-2 Standard",
        ),
        times=1,
    )


@pytest.mark.parametrize("is_semi_configuration", [True, False])
def test_thermocycler_module_conflict_checking(
    decoy: Decoy,
    subject: Deck,
    is_semi_configuration: bool,
) -> None:
    """It should correctly call the deck conflict checker when given a Thermocycler."""
    input_thermocycler = decoy.mock(cls=ThermocyclerGeometry)
    decoy.when(input_thermocycler.highest_z).then_return(42)
    decoy.when(input_thermocycler.load_name).then_return("thermocycler_load_name")
    decoy.when(input_thermocycler.is_semi_configuration).then_return(
        is_semi_configuration
    )

    subject["4"] = input_thermocycler

    decoy.verify(
        deck_conflict.check(
            existing_items={
                DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH,
            },
            new_location=DeckSlotName.SLOT_4,
            new_item=deck_conflict.ThermocyclerModule(
                highest_z_including_labware=42,
                name_for_errors="thermocycler_load_name",
                is_semi_configuration=is_semi_configuration,
            ),
            robot_type="OT-2 Standard",
        ),
        times=1,
    )


def test_heater_shaker_module_conflict_checking(
    decoy: Decoy,
    subject: Deck,
) -> None:
    """It should correctly call the deck conflict checker when given a Heater-Shaker."""
    heater_shaker = decoy.mock(cls=HeaterShakerGeometry)
    decoy.when(heater_shaker.highest_z).then_return(42)
    decoy.when(heater_shaker.load_name).then_return("heater_shaker_load_name")

    subject["4"] = heater_shaker

    decoy.verify(
        deck_conflict.check(
            existing_items={
                DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH,
            },
            new_location=DeckSlotName.SLOT_4,
            new_item=deck_conflict.HeaterShakerModule(
                highest_z_including_labware=42,
                name_for_errors="heater_shaker_load_name",
            ),
            robot_type="OT-2 Standard",
        ),
        times=1,
    )


def test_other_module_conflict_checking(
    decoy: Decoy,
    subject: Deck,
) -> None:
    """It should correctly call the deck conflict checker when given a module."""
    heater_shaker = decoy.mock(cls=ModuleGeometry)
    decoy.when(heater_shaker.highest_z).then_return(42)
    decoy.when(heater_shaker.load_name).then_return("module_load_name")

    subject["4"] = heater_shaker

    decoy.verify(
        deck_conflict.check(
            existing_items={
                DeckSlotName.FIXED_TRASH: EXPECTED_FIXED_TRASH,
            },
            new_location=DeckSlotName.SLOT_4,
            new_item=deck_conflict.OtherModule(
                highest_z_including_labware=42,
                name_for_errors="module_load_name",
            ),
            robot_type="OT-2 Standard",
        ),
        times=1,
    )



"""Test legacy custom labware in an end-to-end environment.

Legacy ProtocolContext objects are prohibitively difficult to instansiate
and mock in an isolated unit test environment.
"""

import textwrap
from pathlib import Path
from typing import List

import pytest
from decoy import matchers

from opentrons_shared_data import load_shared_data

from opentrons.protocol_engine import DeckSlotLocation, LoadedLabware
from opentrons.protocol_reader import ProtocolReader
from opentrons.protocol_runner.create_simulating_orchestrator import (
    create_simulating_orchestrator,
)
from opentrons.types import DeckSlotName

FIXTURE_LABWARE_DEF = load_shared_data("labware/fixtures/2/fixture_96_plate.json")
CUSTOM_LABWARE_PROTOCOL = textwrap.dedent(
    """
    metadata = {
        "apiLevel": "2.11",
    }

    def run(ctx):
        plate = ctx.load_labware(
            load_name="fixture_96_plate",
            location="1",
        )
    """
)


@pytest.fixture()
def custom_labware_protocol_files(tmp_path: Path) -> List[Path]:
    """Create input files to feed into the ProtocolReader."""
    protocol_path = tmp_path / "protocol-name.py"
    protocol_path.write_text(CUSTOM_LABWARE_PROTOCOL)

    labware_path = tmp_path / "labware.json"
    labware_path.write_bytes(FIXTURE_LABWARE_DEF)

    return [protocol_path, labware_path]


async def test_legacy_custom_labware(custom_labware_protocol_files: List[Path]) -> None:
    """It should map legacy pick up tip commands."""
    protocol_reader = ProtocolReader()
    protocol_source = await protocol_reader.read_saved(
        files=custom_labware_protocol_files,
        directory=None,
    )

    subject = await create_simulating_orchestrator(
        robot_type="OT-2 Standard", protocol_config=protocol_source.config
    )
    result = await subject.run(deck_configuration=[], protocol_source=protocol_source)

    expected_labware = LoadedLabware.model_construct(
        id=matchers.Anything(),
        location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        loadName="fixture_96_plate",
        definitionUri="fixture/fixture_96_plate/1",
        offsetId=None,
    )

    assert result.state_summary.errors == []
    assert expected_labware in result.state_summary.labware
    await subject.finish()
